from ..services import dspace
from ..classes import Paper, Corpus

import tempfile
import cPickle as pickle
from unidecode import unidecode

import logging
logging.basicConfig(filename=None, format='%(asctime)-6s: %(name)s - %(levelname)s - %(module)s - %(funcName)s - %(lineno)d - %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel('INFO')

def read_corpus(metapath, bitstreampath=None):
    """
    Generate a :class:`.Corpus` from a ``collection`` in the DSpace repository
    at ``endpoint``.
    
    :class:`.Paper`\s are indexed by URI.
    """
    
    with open(metapath, 'r') as f:
        items = pickle.load(f)
        
    papers = read(metapath, items=items)
    corpus = Corpus(papers, index_by='uri')

    if bitstreampath is not None:
        corpus = add_bitstreams(corpus, bitstreampath)

    return corpus
    
def download_collection(    publickey, privatekey, endpoint, collection, 
                            outpath=None, getBitstreams=False   ):
    """
    Retrieves metadata (and bitstreams) for a ``collection`` in a DSpace
    repository at ``endpoint``.
    
    Parameters
    ----------
    publickey : str
    privatekey : str
    endpoint : str
        REST endpoint.
    collection : int
        Collection ID.
    outpath : str
        (optional). Directory in which to store metadata (and bitstreams).
    getBitstreams : bool
        (default: False) If True, downloads the primary bitstream for each item,
        and stores a dict mapping item URIs onto local paths to those
        bitstreams.
    
    Returns
    -------
    items : list
        A list of items (dict) in the collection.
    bitstreams : dict
        Maps item URIs onto local paths to bitstream files.
    outpath : str
        Path to directory containing metadata (and bitstreams).
    """
    logging.debug('Retrieving metadata from collection {0}'.format(collection))
    
    if outpath is None:
        # Generate a temporary directory to hold metadata
        outpath = tempfile.mkdtemp()    
        logging.debug('Generated temppath at {0}'.format(outpath))
    
    service = dspace.DSpace(publickey, privatekey, endpoint)
    items = service.list_items(collection)
    logging.debug('Retrieved metadata for {0} items'.format(len(items)))    
    
    with open(outpath+'/metadata.pickle', 'w') as f:
        pickle.dump(items, f)
        
    if getBitstreams:
        logging.debug('Retrieving bitstreams from collection {0}'
                                                            .format(collection))
        bitstreams = _get_bitstreams(items, service, outpath)
        # bitstreams is a dict keyed by URI.
    
        with open(outpath+'/bitstreams.pickle', 'w') as f:
            pickle.dump(bitstreams, f)
    else:
        bitstreams = None

    return items, bitstreams, outpath
    
def add_bitstreams(corpus, bitstreampath):
    """
    Load bitstream contents and add them to their respective :class:`.Paper`\s
    in ``corpus``.
    
    Parameters
    ----------
    corpus : :class:`.Corpus`
    bitstreampath : str
        Path to a pickled dict mapping :class:`.Paper` URIs onto paths to files
        containing text data.
        
    Returns
    -------
    corpus : :class:`.Corpus`
    """
    import codecs
    
    with open(bitstreampath, 'r') as f:
        bitstreams = pickle.load(f)
        
    for uri, bpath in bitstreams.items():
        with codecs.open(bpath, 'r', encoding='utf-8') as f:
            contents = unidecode(f.read())

        corpus.papers[uri]['contents'] = contents    
    return corpus

def read(metapath=None, items=None):
    """
    Get metadata from ``collection``, and generate a list of :class:`.Paper`\s.
    """
    
    # Want to be able to provide items to cut down on API calls.
    if items is None:
        with open(metapath, 'r') as f:
            items = pickle.load(f)    
    
    papers = []
    for i in items:
        p = Paper()
        p['uri'] = i['uri']
        p['date'] = int(i['dateCreated'][0:4])
        p['atitle'] = i['title']
        p['abstract'] = i['abstract']
        p['aulast'], p['auinit'] = _handle_creators_string(i['creators_string'])
        p['auuri'] = i['creators']
        papers.append(p)

    logging.debug('Generated {0} papers from metadata'.format(len(papers)))
    return papers

def _get_bitstreams(items, service, outpath):
    bitstreams = {}
    
    # Download bitstreams.
    for i in items:
        if int(i['primary_bitstream']) != -1:
            logger.debug('Getting bitstream for {0}'.format(i['uri']) )
            
            safe_uri = i['uri'].replace('/','-').replace(':','_')
            fpath = '{0}/{1}.txt'.format(outpath, safe_uri)
            bitstream = service.get_bitstream(i['primary_bitstream'], fpath)
            bitstreams[i['uri']] = fpath
                
    return bitstreams

def _handle_creators_string(creators_string):
    """
    Extract author's lastname and initials.
    """
    aulast = []
    auinit = []
    for creator in creators_string:
        try:    # Should be formatted like LAST, FIRST[ MIDDLE][, extra stuff]
            surname, forename = creator.split(', ')[0:2]
        except ValueError: # No comma! Treat as FIRST MIDDLE LAST
            csplit = creator.split()
            surname = csplit[-1]
            forename = ' '.join(csplit[0:-1])
            
        aulast.append(surname.strip().upper())
        
        inits = ''.join([ n[0] for n in forename.split() ])
        auinit.append(inits.strip().upper())
    
    return aulast, auinit
        
    
    
#        print i['creators']

    

#def read_corpus(publickey, privatekey, collectionid):
#    service = dspace.DSpace(publickey, privatekey)
    




import networkx as nx
from itertools import combinations
from collections import Counter, defaultdict

from tethne.utilities import _iterable
from tethne import Corpus, FeatureSet, StructuredFeatureSet


def _generate_graph(graph_class, pairs, node_attrs={}, edge_attrs={},
                    min_weight=1):
    graph = graph_class()
    for combo, count in pairs.items():
        if count >= min_weight:
            if combo in edge_attrs:
                attrs = edge_attrs[combo]
            else:
                attrs = {}
            graph.add_edge(combo[0], combo[1], weight=count, **attrs)

    for k, attrs in node_attrs.items():
        if k in graph.node:
            graph.node[k].update(attrs)
    return graph


def _get_featureset(corpus_or_featureset, featureset_name):
    if type(corpus_or_featureset) is Corpus:  # Retrieve FeatureSet from Corpus.
        if not featureset_name:
            raise ValueError('featureset_name must be provided for Corpus')
        if featureset_name not in corpus_or_featureset.features:
            corpus_or_featureset.index_feature(featureset_name)
        return corpus_or_featureset.features[featureset_name]
    elif type(corpus_or_featureset) in [FeatureSet, StructuredFeatureSet]:
        return corpus_or_featureset     # Already a FeatureSet.
    else:
        raise ValueError('First parameter must be Corpus or FeatureSet')


def cooccurrence(corpus_or_featureset, featureset_name=None, min_weight=1,
                 edge_attrs=['ayjid', 'date'],
                 filter=lambda f, v, c, dc: True):
    """
    A network of feature elements linked by their joint occurrence in papers.
    """

    featureset = _get_featureset(corpus_or_featureset, featureset_name)

    if type(corpus_or_featureset) is Corpus:
        attributes = {i: {a: p[a] for a in edge_attrs}
                      for i, p in corpus_or_featureset.indexed_papers.items()}

    c = lambda f: featureset.count(f)           # Overall count.
    dc = lambda f: featureset.documentCount(f)  # Document count.
    attributes = {}

    # select applies filter to the elements in a (Structured)Feature. The
    #  iteration behavior of Feature and StructuredFeature are different, as is
    #  the manner in which the count for an element in each (Structured)Feature.
    if type(featureset) is FeatureSet:
        select = lambda feature: [f for f, v in feature
                                  if filter(f, v, c(f), dc(f))]
    elif type(featureset) is StructuredFeatureSet:
        select = lambda feature: [f for f in feature
                                  if filter(f, feature.count(f), c(f), dc(f))]

    pairs = Counter()
    eattrs = defaultdict(dict)
    nattrs = defaultdict(dict)
    nset = set()
    for paper, feature in featureset.items():
        if len(feature) == 0:
            continue

        selected = select(feature)
        nset |= set(selected)
        for combo in combinations(selected, 2):
            combo = tuple(sorted(combo))
            pairs[combo] += 1

            if paper in attributes:
                eattrs[combo] = attributes[paper]

    # Generate node attributes.
    for n in list(nset):
        nattrs[n]['count'] = featureset.count(n)
        nattrs[n]['documentCount'] = featureset.documentCount(n)

    graph = _generate_graph(nx.Graph, pairs, edge_attrs=eattrs,
                            node_attrs=nattrs, min_weight=min_weight)

    return graph


def coupling(corpus_or_featureset, featureset_name=None,
             min_weight=1, filter=lambda f, v, c, dc: True):
    """
    A network of papers linked by their joint posession of features.
    """

    featureset = _get_featureset(corpus_or_featureset, featureset_name)

    c = lambda f: featureset.count(f)           # Overall count.
    dc = lambda f: featureset.documentCount(f)  # Document count.
    f = lambda elem: featureset.index[elem]
    v = lambda p, f: featureset.features[p].value(f)

    select = lambda p, elem: filter(f(elem), v(p, f(elem)), c(f(elem)), dc(f(elem)))

    pairs = defaultdict(list)
    for elem, papers in featureset.with_feature.items():
        selected = [p for p in papers if select(p, elem)]
        for combo in combinations(selected, 2):
            combo = tuple(sorted(combo))
            pairs[combo].append(featureset.index[elem])

    graph = nx.Graph()
    for combo, features in pairs.items():
        count = len(features)
        if count >= min_weight:
            graph.add_edge(combo[0], combo[1], features=features, weight=count)
    return graph


def multipartite(corpus, featureset_names, min_weight=1, filters={}):
    """
    A network of papers and one or more featuresets.
    """

    pairs = Counter()
    node_type = {corpus._generate_index(p): {'type': 'paper'}
                 for p in corpus.papers}
    for featureset_name in featureset_names:
        ftypes = {}

        featureset = _get_featureset(corpus, featureset_name)
        for paper, feature in featureset.items():
            if featureset_name in filters:
                if not filters[featureset_name](featureset, feature):
                    continue
            if len(feature) < 1:
                continue
            for f in list(zip(*feature))[0]:
                ftypes[f] = {'type': featureset_name}
                pairs[(paper, f)] += 1
        node_type.update(ftypes)

    return _generate_graph(nx.Graph, pairs, node_attrs=node_type,
                           min_weight=min_weight)



"""
Tethne is a Python package that draws together tools and techniques from
bibliometrics, computational linguistics, and social influence modeling into a
single easy-to-use corpus analysis framework. Scholars can use Tethne to parse
and organize data from the ISI Web of Science and JSTOR Data-for-Research
databases, and generate time-variant citation-based network models, topic
models, and social influence models.

.. autosummary::

   analyze
   classes
   model
   networks
   readers
   writers

"""

from tethne.classes.paper import Paper
from tethne.classes.corpus import Corpus
from tethne.classes.feature import Feature, FeatureSet, \
                                   StructuredFeature, StructuredFeatureSet
from tethne.classes.graphcollection import GraphCollection
from tethne.networks.base import *
from tethne.networks.authors import *
from tethne.networks.papers import *
from tethne.networks.features import *
from tethne.writers.graph import write_graphml, write_csv
from tethne.writers.corpus import write_documents, write_documents_dtm
from tethne.model.corpus.mallet import LDAModel

from tethne.utilities import tokenize, normalize



"""
Write NetworkX graphs to structured and unstructured network file formats.

.. autosummary::

   write_csv
   write_graphml

"""

import networkx as nx
from networkx.readwrite.graphml import GraphMLWriter
import csv
import warnings
from itertools import repeat

from xml.etree.cElementTree import Element, ElementTree, tostring

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


def write_csv(graph, prefix):
    """
    Write ``graph`` as tables of nodes (``prefix-nodes.csv``) and edges
    (``prefix-edges.csv``).

    Parameters
    ----------
    graph : :class:`networkx.Graph`
    prefix : str
    """
    node_headers = list(set([a for n, attrs in graph.nodes(data=True)
                             for a in attrs.keys()]))
    edge_headers = list(set([a for s, t, attrs in graph.edges(data=True)
                             for a in attrs.keys()]))

    value = lambda attrs, h: _recast_value(attrs[h]) if h in attrs else ''
    with open(prefix + '_nodes.csv', 'w') as f:
        writer = csv.writer(f)
        writer.writerow(['node'] + node_headers)
        for n, attrs in graph.nodes(data=True):
            values = map(value, repeat(attrs, len(node_headers)), node_headers)
            writer.writerow([_recast_value(n)] + list(values))

    with open(prefix + '_edges.csv', 'w') as f:
        writer = csv.writer(f)
        writer.writerow(['source', 'target'] + edge_headers)
        for s, t, attrs in graph.edges(data=True):
            values = map(value, repeat(attrs, len(edge_headers)), edge_headers)
            writer.writerow([_recast_value(s), _recast_value(t)] + list(values))


def to_sif(graph, output_path):
    """
    Generates Simple Interaction Format output file from provided graph.

    The SIF specification is described
    `here <http://wiki.cytoscape.org/Cytoscape_User_Manual/Network_Formats>`_.

    :func:`.to_sif` will generate a .sif file describing the network, and a few
    .eda and .noa files containing edge and node attributes, respectively. These
    are equivalent to tab-delimited tables, and can be imported as such in
    Cytoscape 3.0.

    Parameters
    ----------
    graph : networkx.Graph
        The Graph to be exported to SIF.
    output_path : str
        Full path, including filename (without suffix).
        e.g. using "./graphFolder/graphFile" will result in a SIF file at
        ./graphFolder/graphFile.sif, and corresponding .eda and .noa files.

    """

    warnings.warn("Removed in 0.8. Use write_csv instead.", DeprecationWarning)

    graph = _strip_list_attributes(graph)

    if output_path[-4:] == ".sif":
        output_path = output_path[:-4]

    if nx.number_of_nodes(graph) == 0:
        # write an empty file (for a non-existant graph)
        f = open(output_path + ".sif","w")
        f.write("")
        f.close()

    else:
        # write node files
        nodes = graph.nodes(data=True)
        for node in nodes:
            node_name = node[0]
            node_attribs = node[1]
            for key, value in node_attribs.items():
                # generate a node attribute file for each node attribute
                if node == nodes[0]:
                    # first node, overwrite file
                    with open(output_path + "_" + key + ".noa",
                              "w") as f:
                        f.write(unicode(key) + '\n')
                        f.write(unicode(node_name).replace(" ","_") + " = " +
                                unicode(value) + "\n")
                else:
                    # not first, append file
                    with open(output_path + "_" + key + ".noa",
                              "a") as f:
                        f.write(unicode(node_name).replace(" ","_") + " = " +
                                unicode(value) + "\n")

        if nx.number_of_edges(graph) == 0:
            # write an empty graph to a .sif file (just its nodes)
            for node in nodes:
                node_name = node[0]
                if node == nodes[0]:
                    # first node, overwrite file
                    with open(output_path + ".sif","w") as f:
                        f.write(unicode(node_name).replace(" ","_") + "\n")
                else:
                    # not first, append file
                    with open(output_path + ".sif","a") as f:
                        f.write(unicode(node_name).replace(" ","_") + "\n")

        else:
            # write the graph to a .sif file as well as other edge
            # attribute files

            if graph.is_multigraph():
                # then the NetworkX graph supports multiple interaction
                # types just like the .sif format
                edges = graph.edges(data=True, keys=True)
                edge_attribs = set()
                for edge in edges:
                    for key in edge[3].iterkeys():
                        edge_attribs.add(key)

                # create edge attribute files
                for attrib in edge_attribs:
                    str_attrib = unicode(attrib)
                    with open(output_path + '_' + str_attrib + ".eda","w") as f:
                        f.write(unicode(attrib) + "\n")

                # add data to eda files and write sif file
                with open(output_path + '.sif', 'w') as f:
                    for edge in edges:
                        node1 = unicode(edge[0]).replace(" ", "_")
                        node2 = unicode(edge[1]).replace(" ", "_")
                        intr_type = unicode(edge[2]).replace(" ", "_")
                        sif_line = node1 + ' ' + intr_type + ' ' + node2 + '\n'
                        f.write(sif_line)

                        for attrib, value in edge[3].items():
                            eda_line = (node1 + ' (' + intr_type + ') ' +
                                        node2 + ' = ' + unicode(value) + '\n')
                            with open(output_path + '_' + unicode(attrib) + '.eda',
                                      'a') as g:
                                g.write(eda_line)

            else:
                # then we support only one interaction type 'rel'
                edges = graph.edges(data=True)
                edge_attribs = set()
                for edge in edges:
                    for key in edge[2].iterkeys():
                        edge_attribs.add(key)

                # create edge attribute files
                for attrib in edge_attribs:
                    str_attrib = unicode(attrib)
                    with open(output_path + '_' + str_attrib + ".eda","w") as f:
                        f.write(unicode(attrib) + "\n")

                # add data to eda files and write sif file
                with open(output_path + '.sif', 'w') as f:
                    for edge in edges:
                        node1 = unicode(edge[0]).replace(" ", "_")
                        node2 = unicode(edge[1]).replace(" ", "_")
                        intr_type = 'rel'
                        sif_line = node1 + ' ' + intr_type + ' ' + node2 + '\n'
                        f.write(sif_line)

                        for attrib, value in edge[2].items():
                            eda_line = (node1 + ' (' + intr_type + ') ' +
                                        node2 + ' = ' + unicode(value) + '\n')
                            with open(output_path + '_' + unicode(attrib) +
                                      '.eda', 'a') as g:
                                g.write(eda_line)


def to_gexf(graph, output_path):
    """Writes graph to `GEXF <http://gexf.net>`_.

    Uses the NetworkX method
    `write_gexf <http://networkx.lanl.gov/reference/generated/networkx.readwrite.gexf.write_gexf.html>`_.

    Parameters
    ----------
    graph : networkx.Graph
        The Graph to be exported to GEXF.
    output_path : str
        Full path, including filename (without suffix).
        e.g. using "./graphFolder/graphFile" will result in a GEXF file at
        ./graphFolder/graphFile.gexf.
    """
    warnings.warn("Removed in 0.8.", DeprecationWarning)

    graph = _strip_list_attributes(graph)

    nx.write_gexf(graph, output_path + ".gexf")


def write_graphml(graph, path, encoding='utf-8', prettyprint=True):
    """Writes graph to `GraphML <http://graphml.graphdrawing.org/>`_.

    Uses the NetworkX method
    `write_graphml <http://networkx.lanl.gov/reference/generated/networkx.readwrite.graphml.write_graphml.html>`_.

    Parameters
    ----------
    graph : networkx.Graph
        The Graph to be exported to GraphML.
    output_path : str
        Full path, including filename (without suffix).
        e.g. using "./graphFolder/graphFile" will result in a GraphML file at
        ./graphFolder/graphFile.graphml.
    """
    graph = _strip_list_attributes(graph)

    writer = TethneGraphMLWriter(encoding=encoding, prettyprint=prettyprint)
    writer.add_graph_element(graph)
    writer.dump(open(path, 'wb'))


def to_graphml(graph, path, encoding='utf-8', prettyprint=True):
    warnings.warn("Removed in 0.8. Use write_graphml instead.",
                  DeprecationWarning)

    write_graphml(graph, path, encoding=encoding, prettyprint=prettyprint)


class TethneGraphMLWriter(GraphMLWriter):
    def get_key(self, name, attr_type, scope, default):
        # Modified to use attribute name as key, rather than numeric ID.
        keys_key = (name, attr_type, scope)
        try:
            return self.keys[keys_key]
        except KeyError:
            new_id = name
            self.keys[keys_key] = new_id
            key_kwargs = {"id": new_id,
                          "for": scope,
                          "attr.name": name,
                          "attr.type": attr_type}
            key_element = Element("key", **key_kwargs)
            # add subelement for data default value if present
            if default is not None:
                default_element = Element("default")
                default_element.text = _recast_value(default)
                key_element.append(default_element)
            self.xml.insert(0, key_element)
        return new_id

    def add_data(self, name, e_type, value, scope="all", default=None):
        if e_type not in self.xml_type:
            raise nx.NetworkXError('GraphML writer does not support '
                                   '%s as data values.'%e_type)
        key_id = self.get_key(name, self.xml_type[e_type], scope, default)
        data_element = Element("data", key=key_id)
        data_element.text = _recast_value(value)
        return data_element

    def add_nodes(self, G, graph_element):
        for node,data in G.nodes_iter(data=True):
            node_element = Element("node", id=_recast_value(node))
            default = G.graph.get('node_default', {})
            self.add_attributes("node", node_element, data, default)
            graph_element.append(node_element)

    def add_edges(self, G, graph_element):
        if G.is_multigraph():
            for u,v,key,data in G.edges_iter(data=True, keys=True):
                edge_element = Element("edge",source=_recast_value(u),
                                       target=_recast_value(v))
                default = G.graph.get('edge_default',{})
                self.add_attributes("edge", edge_element, data, default)
                self.add_attributes("edge", edge_element, {'key': key},
                                    default)
                graph_element.append(edge_element)
        else:
            for u,v,data in G.edges_iter(data=True):
                edge_element = Element("edge", source=_recast_value(u),
                                       target=_recast_value(v))
                default = G.graph.get('edge_default', {})
                self.add_attributes("edge", edge_element, data, default)
                graph_element.append(edge_element)


def to_table(graph, path):
    warnings.warn("Removed in 0.8. Use write_csv instead.",
                  DeprecationWarning)

    graph = _strip_list_attributes(graph)

    # Edge list.
    with open(path + "_edges.csv", "wb") as f:
        edges = graph.edges(data=True)
        writer = csv.writer(f, delimiter='\t')

        # Header.
        writer.writerow(['source','target'] + [ k for k in edges[0][2].keys()])

        # Values.
        for e in edges:
            writer.writerow([ e[0], e[1]] + [ v for v in e[2].values() ] )

    # Node attributes.
    with open(path + "_nodes.csv", "wb") as f:
        nodes = graph.nodes(data=True)
        writer = csv.writer(f, delimiter='\t')

        # Header.
        writer.writerow(['node'] + [ k for k in nodes[0][1].keys() ])

        # Values.
        for n in nodes:
            writer.writerow([ n[0] ] + [ v for v in n[1].values() ])


def _strip_list_attributes(G):
    for n in G.nodes(data=True):
        for k,v in n[1].items():
            if type(v) is list:
                G.node[n[0]][k] = unicode(v)
    for e in G.edges(data=True):
        for k,v in e[2].items():
            if type(v) is list:
                G.edge[e[0]][e[1]][k] = unicode(v)

    return G


def _recast_value(value):
    if type(value) in [str, int, unicode, float]:
        return unicode(value)
    if hasattr(value, '__iter__'):
        return ', '.join(list(value))



"""
A :class:`.Paper` represents a single bibliographic record.
"""

from tethne.classes.feature import Feature, feature

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


class Paper(object):
    """
    Tethne's representation of a bibliographic record.

    Fields can be set using dict-like assignment, and accessed as attributes.

    .. code-block:: python

       >>> myPaper = Paper()
       >>> myPaper['date'] = 1965
       >>> myPaper.date
       1965

    """

    def __setitem__(self, key, value):
        if hasattr(self, key):  # Don't mess with class methods.
            if hasattr(getattr(self, key), '__call__'):
                return
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    @property
    def ayjid(self):
        """
        Fuzzy WoS-style identifier, comprised of first author's name (LAST I),
        pubdate, and journal.

        Returns
        -------
        ayjid : str
        """

        try:
            return self._ayjid
        except:
            pass

        if hasattr(self, 'authors_init'):
                al, ai = self.authors_init[0]
        elif hasattr(self, 'authors_full'):
                al, af = self.authors_init[0]
                ai = u''.join([i[0] for i in af.split(' ')])
        else:
            al, ai = u'', u''
        if hasattr(self, 'date'):
            date = unicode(self.date)
        else:
            date = u''
        if hasattr(self, 'journal'):
            journal = self.journal
        else:
            journal = u''

        ayjid = u' '.join([al, ai.replace(' ', ''), date, journal]).strip()
        self._ayjid = ayjid
        return ayjid.upper()


    @property
    @feature
    def authors(self):
        """
        Get the authors of the current :class:`.Paper` instance.

        Uses ``authors_full`` if it is available. Otherwise, uses
        ``authors_init``.

        Returns
        -------
        authors : :class:`.Feature`
            Author names are in the format ``LAST F``.
        """

        if hasattr(self, 'authors_full'):
            return self.authors_full
        elif hasattr(self, 'authors_init'):
            return self.authors_init
        else:
            return []


    @property
    @feature
    def citations(self):
        """
        Cited references as a :class:`.Feature`\.

        Returns
        -------
        citations : :class:`.Feature`
        """

        if hasattr(self, 'citedReferences'):
            return [cr.ayjid for cr in self.citedReferences if cr is not None]
        return []



"""
Methods for analyzing :class:`.Corpus` objects.

.. autosummary::
   :nosignatures:

   burstness
   feature_burstness
   sigma

"""

import networkx as nx
import warnings

from math import exp, log
from collections import defaultdict
from itertools import izip

from tethne.utilities import argmin, mean

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str
    xrange = range



def _forward(X, s=1.1, gamma=1., k=5):
    """
    Forward dynamic algorithm for burstness automaton HMM, from `Kleinberg
    (2002) <http://www.cs.cornell.edu/home/kleinber/bhs.pdf>`_.

    Parameters
    ----------
    X : list
        A series of time-gaps between events.
    s : float
        (default: 1.1) Scaling parameter ( > 1.)that controls graininess of
        burst detection. Lower values make the model more sensitive.
    gamma : float
        (default: 1.0) Parameter that controls the 'cost' of higher burst
        states. Higher values make it more 'difficult' to achieve a higher
        burst state.
    k : int
        (default: 5) Number of states. Higher values increase computational
        cost of the algorithm. A maximum of 25 is suggested by the literature.

    Returns
    -------
    states : list
        Optimal state sequence.
    """
    X = list(X)

    def alpha(i):
        return (n/T)*(s**i)

    def tau(i, j):
        if j > i:
            return (j-i)*gamma*log(n)
        return 0.

    def f(j, x):
        return alpha(j) * exp(-1. * alpha(j) * x)

    def C(j, t):
        if j == 0 and t == 0:
            return 0.
        elif t == 0:
            return float("inf")

        C_tau = min([C_values[l][t-1] + tau(l, j) for l in xrange(k)])
        return (-1. * log(f(j,X[t]))) + C_tau

    T = sum(X)
    n = len(X)

    # C() requires default (0) values, so we construct the "array" in advance.
    C_values = [[0 for t in xrange(len(X))] for j in xrange(k)]
    for j in xrange(k):
        for t in xrange(len(X)):
            C_values[j][t] = C(j,t)

    # Find the optimal state sequence.
    states = [argmin([c[t] for c in C_values]) for t in xrange(n)]
    return states

def _top_features(corpus, feature, topn=20, perslice=False, axis='date'):
    warnings.warn("Removed in 0.8. Use corpus.top_features() instead.",
                  DeprecationWarning)
    return corpus.top_features(feature, topn=topn, perslice=perslice)

def burstness(corpus, featureset_name, features=[], k=5, topn=20,
              perslice=False, normalize=True, **kwargs):
    """
    Estimate burstness profile for the ``topn`` features (or ``flist``) in
    ``feature``.

    Uses the popular burstness automaton model inroduced by `Kleinberg (2002)
    <http://www.cs.cornell.edu/home/kleinber/bhs.pdf>`_.

    Parameters
    ----------
    corpus : :class:`.Corpus`
    feature : str
        Name of featureset in ``corpus``. E.g. ``'citations'``.
    k : int
        (default: 5) Number of burst states.
    topn : int or float {0.-1.}
        (default: 20) Number (int) or percentage (float) of top-occurring
        features to return. If ``flist`` is provided, this parameter is ignored.
    perslice : bool
        (default: False) If True, loads ``topn`` features per slice. Otherwise,
        loads ``topn`` features overall. If ``flist`` is provided, this
        parameter is ignored.
    flist : list
        List of features. If provided, ``topn`` and ``perslice`` are ignored.
    normalize : bool
        (default: True) If True, burstness is expressed relative to the hightest
        possible state (``k-1``). Otherwise, states themselves are returned.
    kwargs : kwargs
        Parameters for burstness automaton HMM.

    Returns
    -------
    B : dict
        Keys are features, values are tuples of ( dates, burstness )

    Examples
    --------

    .. code-block:: python

       >>> from tethne.analyze.corpus import burstness
       >>> B = burstness(corpus, 'abstractTerms', flist=['process', 'method']
       >>> B['process']
       ([1990, 1991, 1992, 1993], [0., 0.4, 0.6, 0.])

    """

    # If `features` of interest are not specified, calculate burstness for the
    #  top `topn` features.
    if len(features) == 0:
        T = corpus.top_features(featureset_name, topn=topn, perslice=perslice)
        if perslice:
            features = list(set([f[0] for flist in zip(*T)[1] for f in flist]))
        else:
            features = list(zip(*T))[0]

    B = {feature: feature_burstness(corpus, featureset_name, feature, k=k,
                                    normalize=normalize, **kwargs)
         for feature in features}
    return B

def feature_burstness(corpus, featureset_name, feature, k=5, normalize=True,
                      s=1.1, gamma=1., **slice_kwargs):
    """
    Estimate burstness profile for a feature over the ``'date'`` axis.

    Parameters
    ----------
    corpus : :class:`.Corpus`
    feature : str
        Name of featureset in ``corpus``. E.g. ``'citations'``.
    findex : int
        Index of ``feature`` in ``corpus``.
    k : int
        (default: 5) Number of burst states.
    normalize : bool
        (default: True) If True, burstness is expressed relative to the hightest
        possible state (``k-1``). Otherwise, states themselves are returned.
    kwargs : kwargs
        Parameters for burstness automaton HMM.
    """


    if featureset_name not in corpus.features:
        corpus.index_feature(featureset_name)

    if 'date' not in corpus.indices:
        corpus.index('date')

    # Get time-intervals between occurrences.
    dates = [min(corpus.indices['date'].keys()) - 1]    # Pad start.
    X_ = [1.]

    years, values = corpus.feature_distribution(featureset_name, feature)
    for year, N in izip(years, values):
        if N == 0:
            continue

        if N > 1:
            if year == dates[-1] + 1:
                for n in xrange(int(N)):
                    X_.append(1./N)
                    dates.append(year)
            else:
                X_.append(float(year - dates[-1]))
                dates.append(year)
                for n in xrange(int(N) - 1):
                    X_.append(1./(N - 1))
                    dates.append(year)
        else:
            X_.append(float(year - dates[-1]))
            dates.append(year)

    # Get optimum state sequence.
    st = _forward(map(lambda x: x*100, X_), s=s, gamma=gamma, k=k)

    # Bin by date.
    A = defaultdict(list)
    for i in xrange(len(X_)):
        A[dates[i]].append(st[i])

    # Normalize.
    if normalize:
        A = {key: mean(values)/k for key, values in A.items()}
    else:
        A = {key: mean(values) for key, values in A.items()}

    D = sorted(A.keys())
    return D[1:], [A[d] for d in D[1:]]


def sigma(G, corpus, featureset_name, B=None, **kwargs):
    """
    Calculate sigma (from `Chen 2009 <http://arxiv.org/pdf/0904.1439.pdf>`_)
    for all of the nodes in a :class:`.GraphCollection`\.

    You can set parameters for burstness estimation using ``kwargs``:

    =========   ===============================================================
    Parameter   Description
    =========   ===============================================================
    s           Scaling parameter ( > 1.)that controls graininess of burst
                detection. Lower values make the model more sensitive. Defaults
                to 1.1.
    gamma       Parameter that controls the 'cost' of higher burst states.
                Defaults to 1.0.
    k           Number of burst states. Defaults to 5.
    =========   ===============================================================

    Parameters
    ----------
    G : :class:`.GraphCollection`
    corpus : :class:`.Corpus`
    feature : str
        Name of a featureset in `corpus`.


    Examples
    --------

    Assuming that you have a :class:`.Corpus` generated from WoS data that has
    been sliced by ``date``.

    .. code-block:: python

       >>> # Generate a co-citation graph collection.
       >>> from tethne import GraphCollection
       >>> kwargs = { 'threshold':2, 'topn':100 }
       >>> G = GraphCollection()
       >>> G.build(corpus, 'date', 'papers', 'cocitation', method_kwargs=kwargs)

       >>> # Calculate sigma. This may take several minutes, depending on the
       >>> #  size of your co-citaiton graph collection.
       >>> from tethne.analyze.corpus import sigma
       >>> G = sigma(G, corpus, 'citations')

       >>> # Visualize...
       >>> from tethne.writers import collection
       >>> collection.to_dxgmml(G, '~/cocitation.xgmml')

    In the visualization below, node and label sizes are mapped to ``sigma``,
    and border width is mapped to ``citations``.

    .. figure:: _static/images/cocitation_sigma2.png
       :width: 600
       :align: center

    """
    if 'date' not in corpus.indices:
        corpus.index('date')

    # Calculate burstness if not provided.
    if not B:
        B = burstness(corpus, featureset_name, features=G.nodes(), **kwargs)

    Sigma = {}      # Keys are dates (from GraphCollection), values are
                    #  node:sigma dicts.
    for key, graph in G.items():
        centrality = nx.betweenness_centrality(graph)
        sigma = {}  # Sigma values for all features in this year.
        attrs = {}  # Sigma values for only those features in this graph.
        for n_, burst in B.iteritems():
            burst = dict(list(zip(*burst)))     # Reorganize for easier lookup.

            # Nodes are indexed as integers in the GraphCollection.
            n = G.node_lookup[n_]

            # We have burstness values for years in which the feature ``n``
            #  occurs, and we have centrality values for years in which ``n``
            #  made it into the graph.
            if n in graph.nodes() and key in burst:
                sigma[n] = ((centrality[n] + 1.) ** burst[key]) - 1.
                attrs[n] = sigma[n]

        # Update graph with sigma values.
        nx.set_node_attributes(graph, 'sigma', attrs)
        Sigma[key] = sigma

    # Invert results and update the GraphCollection.master_graph.
    # TODO: is there a more efficient way to do this?
    inverse = defaultdict(dict)
    for gname, result in Sigma.items():
        if hasattr(result, '__iter__'):
            for n, val in result.items():
                inverse[n].update({gname: val})
    nx.set_node_attributes(G.master_graph, 'sigma', inverse)

    # We want to return results in the same format as burstness(); with node
    #  labels as keys; values are tuples ([years...], [sigma...]).
    return {n: list(zip(*G.node_history(G.node_lookup[n], 'sigma').items()))
            for n in B.keys()}



"""
"""

from collections import Counter
from itertools import repeat
import codecs
import os
import csv

from tethne import FeatureSet, StructuredFeatureSet

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


def write_documents(corpus, target, featureset_name, metadata_fields=[]):
    """
    Parameters
    ----------


    """

    docpath = target + '_docs.txt'
    metapath = target + '_meta.csv'

    features = corpus.features[featureset_name].features
    ftype = type(corpus.features[featureset_name])
    index = corpus.features[featureset_name].index

    try:
        docFile = open(docpath, 'wb')
    except IOError:
        raise IOError('Invalid target. Could not open files for writing.')

    # Generate metadata.
    with codecs.open(metapath, 'w', encoding='utf-8') as f:
        writer = csv.writer(f)
        writer.writerow([corpus.index_by] + list(metadata_fields))
        for i, p in corpus.indexed_papers.items():
            getter = lambda m: getattr(p, m) if hasattr(p, m) else None
            writer.writerow([i] + list(map(getter, metadata_fields)))

    # Write documents content.
    with codecs.open(docpath, 'w', encoding='utf-8') as f:
        for i, p in corpus.indexed_papers.items():
            if i in features:
                row = [i, u'en']
                if ftype is FeatureSet:
                    row += [u' '.join(repeat(e, c)) for e, c in features[i]]
                elif ftype is StructuredFeatureSet:
                    row += features[i]
                f.write(u'\t'.join(row) + u'\n')

    return docpath, metapath


def write_documents_dtm(corpus, target, featureset_name, slice_kwargs={},
                        metadata_fields=['date','title']):
    """

    Parameters
    ----------
    target : str
        Target path for documents; e.g. './mycorpus' will result in
        './mycorpus-mult.dat', './mycorpus-seq.dat', 'mycorpus-vocab.dat', and
        './mycorpus-meta.dat'.
    D : :class:`.Corpus`
        Contains :class:`.Paper` objects generated from the same DfR dataset
        as t_ngrams, indexed by doi and sliced by date.
    feature : str
        (default: 'unigrams') Features in :class:`.Corpus` to use for
        modeling.
    fields : list
        (optional) Fields in :class:`.Paper` to include in the metadata file.

    Returns
    -------
    None : If all goes well.

    Raises
    ------
    IOError
    """

    metapath = target + '-meta.dat'
    multpath = target + '-mult.dat'
    seqpath = target + '-seq.dat'
    vpath = target + '-vocab.dat'

    lookup = corpus.features[featureset_name].lookup
    index = corpus.features[featureset_name].index
    features = corpus.features[featureset_name].features

    # Generate -mult.dat file (wordcounts for each document).
    #   From the DTM example:
    #
    #     one-doc-per-line, each line of the form
    #         unique_word_count index1:count1 index2:count2 ... indexn:counnt
    #     The docs in foo-mult.dat should be ordered by date, with the first
    #     docs from time1, the next from time2, ..., and the last docs from
    #     timen.
    #
    N = Counter()
    for date, subcorpus in corpus.slice(**slice_kwargs):
        with codecs.open(multpath, 'w', encoding='utf-8') as f:
            for p in subcorpus.papers:
                i = getattr(p, subcorpus.index_by)
                N[date] += 1
                docLine = [u':'.join([unicode(lookup[e]), unicode(c)])
                           for e,c in features[i]]
                unique = unicode(len(features[i]))
                f.write(u' '.join([unique] + docLine) + '\n')

    # And -meta.dat file (with DOIs).
    #
    #       a file with information on each of the documents, arranged in
    #           the same order as the docs in the mult file.
    #
    with codecs.open(metapath, 'w', encoding='utf-8') as f:
        writer = csv.writer(f, delimiter='\t')
        writer.writerow(['id'] + list(metadata_fields))
        for date, subcorpus in corpus.slice(**slice_kwargs):
            for p in subcorpus.papers:
                getter = lambda m: getattr(p, m) if hasattr(p, m) else None
                fieldData = map(getter, metadata_fields)
                writer.writerow([getattr(p, corpus.index_by)] + list(fieldData))

    # Generate -seq.dat file (number of papers per year).
    #   From the DTM example:
    #
    #       Number_Timestamps
    #       number_docs_time_1
    #       ...
    #       number_docs_time_i
    #       ...
    #       number_docs_time_NumberTimestamps
    #
    with open(seqpath, 'w') as f:
        for date in sorted(N.keys()):
            f.write(u'{date}\n'.format(date=N[date]))

    #       a file with all of the words in the vocabulary, arranged in
    #       the same order as the word indices
    with codecs.open(vpath, 'w', encoding='utf-8') as f:
        f.write(u'\n'.join([index[i] for i in sorted(index.keys())]))



"""
Classes and methods related to the :class:`.MALLETModelManager`\.
"""

import os
import re
import shutil
import tempfile
import subprocess
import csv
import platform
from collections import defaultdict

from networkx import Graph

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str

import logging
logging.basicConfig()
logger = logging.getLogger('mallet')
logger.setLevel('ERROR')

from tethne import write_documents, Feature, FeatureSet
from tethne.model import Model
import tethne

# Determine path to MALLET.
TETHNE_PATH = os.path.split(os.path.abspath(tethne.__file__))[0]
MALLET_PATH = os.path.join(TETHNE_PATH, 'bin', 'mallet-2.0.7')

import sys
if sys.version_info[0] > 2:
    xrange = range


class LDAModel(Model):
    """
    Generates a :class:`.LDAModel` from a :class:`.Corpus` using
    `MALLET <http://mallet.cs.umass.edu/>`_.

    The :class:`.Corpus` should already contain at least one featurset,
    indicated by the `feature` parameter, such as wordcounts. You may
    specify two working directories: `temppath` should be a working
    directory that will contain intermediate files (e.g. documents, data
    files, metadata), while `outpath` will contain the final model and any
    plots generated during the modeling process. If `temppath` is not
    provided, generates and uses a system temporary directory.

    Tethne comes bundled with a recent version of MALLET. If you would
    rather use your own install, you can do so by providing the
    `mallet_path` parameter. This should point to the directory containing
    ``/bin/mallet``.

    .. autosummary::
       :nosignatures:

       topic_over_time

    Parameters
    ----------
    D : :class:`.Corpus`
    feature : str
        Key from D.features containing wordcounts (or whatever
        you want to model with).
    outpath : str
        Path to output directory.
    temppath : str
        Path to temporary directory.
    mallet_path : str
        Path to MALLET install directory (contains bin/mallet).

    Examples
    --------

    Starting with some JSTOR DfR data (with wordcounts), a typical workflow
    might look something like this:

    .. code-block:: python

       >>> from nltk.corpus import stopwords                 #  1. Get stoplist.
       >>> stoplist = stopwords.words()

       >>> from tethne.readers import dfr                    #  2. Build Corpus.
       >>> C = dfr.corpus_from_dir('/path/to/DfR/datasets', 'uni', stoplist)

       >>> def filt(s, C, DC):                           # 3. Filter wordcounts.
       ...     if C > 3 and DC > 1 and len(s) > 3:
       ...         return True
       ...     return False
       >>> C.filter_features('wordcounts', 'wc_filtered', filt)

       >>> from tethne.model import MALLETModelManager       #   4. Get Manager.
       >>> outpath = '/path/to/my/working/directory'
       >>> mallet = '/Applications/mallet-2.0.7'
       >>> M = MALLETModelManager(C, 'wc_filtered', outpath, mallet_path=mallet)

       >>> M.prep()                                          #    5. Prep model.

       >>> model = M.build(Z=50, max_iter=300)               #   6. Build model.
       >>> model                                             # (may take awhile)
       <tethne.model.corpus.ldamodel.LDAModel at 0x10bfac710>

    A plot showing the log-likelihood/topic over modeling iterations should be
    generated in your `outpath`. For example:

    .. figure:: _static/images/ldamodel_LL.png
       :width: 400
       :align: center

    Behind the scenes, the :func:`.prep` procedure generates a plain-text corpus
    file at `temppath`, along with a metadata file. MALLET's ``import-file``
    procedure is then called, which translates the corpus into MALLET's internal
    format (also stored at the `temppath`).

    The :func:`.build` procedure then invokes MALLET's ``train-topics``
    procedure. This step may take a considerable amount of time, anywhere from
    a few minutes (small corpus, few topics) to a few hours (large corpus, many
    topics).

    For a :class:`.Corpus` with a few thousand :class:`.Paper`\s, 300 - 500
    iterations is often sufficient to achieve convergence for 20-100 topics.

    Once the :class:`.LDAModel` is built, you can access its methods directly.
    See full method descriptions in :class:`.LDAModel`\.

    For more information about topic modeling with MALLET see
    `this tutorial <http://programminghistorian.org/lessons/topic-modeling-and-mallet>`_.

    """

    mallet_path = MALLET_PATH

    def __init__(self, *args, **kwargs):
        self.mallet_bin = os.path.join(self.mallet_path, "bin", "mallet")

        if platform.system() == 'Windows':
            self.mallet_bin += '.bat'
        os.putenv('MALLET_HOME', self.mallet_path)
        super(LDAModel, self).__init__(*args, **kwargs)

    def prep(self):
        self.dt = os.path.join(self.temp, "dt.dat")
        self.wt = os.path.join(self.temp, "wt.dat")
        self.om = os.path.join(self.temp, "model.mallet")

        self._generate_corpus()

    def _generate_corpus(self):
        """
        Writes a corpus to disk amenable to MALLET topic modeling.
        """

        target = self.temp + 'mallet'
        paths = write_documents(self.corpus, target, self.featureset_name,
                                ['date', 'title'])
        self.corpus_path, self.metapath = paths

        self._export_corpus()

    def _export_corpus(self):
        """
        Calls MALLET's `import-file` method.
        """
        # bin/mallet import-file --input /Users/erickpeirson/mycorpus_docs.txt
        #     --output mytopic-input.mallet --keep-sequence --remove-stopwords

        if not os.path.exists(self.mallet_bin):
            raise IOError("MALLET path invalid or non-existent.")
        self.input_path = os.path.join(self.temp, "input.mallet")

        exit = subprocess.call([
                self.mallet_bin,
                'import-file',
                '--input', self.corpus_path,
                '--output', self.input_path,
                '--keep-sequence',          # Required for LDA.
                '--remove-stopwords'])      # Probably redundant.

        if exit != 0:
            msg = "MALLET import-file failed with exit code {0}.".format(exit)
            raise RuntimeError(msg)

    def run(self, **kwargs):
        """
        Calls MALLET's `train-topic` method.
        """
        #$ bin/mallet train-topics --input mytopic-input.mallet
        #> --num-topics 100
        #> --output-doc-topics /Users/erickpeirson/doc_top
        #> --word-topic-counts-file /Users/erickpeirson/word_top
        #> --output-topic-keys /Users/erickpeirson/topic_keys

        if not os.path.exists(self.mallet_bin):
            raise IOError("MALLET path invalid or non-existent.")

        for attr in ['Z', 'max_iter']:
            if not hasattr(self, attr):
                raise AttributeError('Please set {0}'.format(attr))

        self.ll = []
        self.num_iters = 0
        logger.debug('run() with k={0} for {1} iterations'.format(self.Z, self.max_iter))

        prog = re.compile(u'\<([^\)]+)\>')
        ll_prog = re.compile(r'(\d+)')
        p = subprocess.Popen([
                    self.mallet_bin,
                    'train-topics',
                    '--input', self.input_path,
                    '--num-topics', unicode(self.Z),
                    '--num-iterations', unicode(self.max_iter),
                    '--output-doc-topics', self.dt,
                    '--word-topic-counts-file', self.wt,
                    '--output-model', self.om],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)

        # Handle output of MALLET in real time.
        while p.poll() is None:
            l = p.stderr.readline()

            # Keep track of LL/topic.
            try:
                this_ll = float(re.findall(u'([-+]\d+\.\d+)', l)[0])
                self.ll.append(this_ll)
            except IndexError:  # Not every line will match.
                pass

            # Keep track of modeling progress.
            try:
                this_iter = float(prog.match(l).groups()[0])
                progress = int(100. * this_iter/self.max_iter)
                print 'Modeling progress: {0}%.\r'.format(progress),
            except AttributeError:  # Not every line will match.
                pass

        self.num_iters += self.max_iter
        self.load()

    def load(self, **kwargs):
        self._read_theta(kwargs.get('dt', self.dt))
        self._read_phi(kwargs.get('wt', self.wt))

    def _read_theta(self, dt):
        """
        Used by :func:`.from_mallet` to reconstruct theta posterior
        distributions.

        Returns
        -------
        td : Numpy array
            Rows are documents, columns are topics. Rows sum to ~1.
        """

        self.theta = FeatureSet()

        with open(dt, "rb") as f:
            i = -1
            reader = csv.reader(f, delimiter='\t')
            for line in reader:
                i += 1
                if i == 0:
                    continue     # Avoid header row.

                d, id, t = int(line[0]), unicode(line[1]), line[2:]
                feature = Feature([(int(t[i]), float(t[i + 1]))
                                   for i in xrange(0, len(t) - 1, 2)])
                self.theta.add(id, feature)

        self.corpus.features['topics'] = self.theta
        return self.theta

    def _read_phi(self, wt):
        """
        Used by :func:`.from_mallet` to reconstruct phi posterior distributions.

        Returns
        -------
        wt : Numpy array
            Rows are topics, columns are words. Rows sum to ~1.
        """

        self.vocabulary = {}
        phi_features = {}

        # TODO: make this encoding-safe.
        with open(wt, "r") as f:
            reader = csv.reader(f, delimiter=' ')
            topics = defaultdict(list)
            for line in reader:
                w, term = int(line[0]), unicode(line[1])
                self.vocabulary[w] = term

                for l in line[2:]:
                    k, c = l.split(':')    # Topic and assignment count.
                    topics[int(k)].append((w, int(c)))

        for k, data in topics.items():
            nfeature = Feature(data).norm

            phi_features[k] = nfeature
        self.phi = FeatureSet(phi_features)

    def topics_in(self, d, topn=5):
        """
        List the top ``topn`` topics in document ``d``.
        """
        return self.theta.features[d].top(topn)

    def list_topic(self, k, Nwords=10):
        """
        List the top ``topn`` words for topic ``k``.


        Examples
        --------

        .. code-block:: python

           >>> model.list_topic(1, Nwords=5)
           [ 'opposed', 'terminates', 'trichinosis', 'cistus', 'acaule' ]

        """

        return [(self.vocabulary[w], p) for w, p
                in self.phi.features[k].top(Nwords)]

    def list_topics(self, Nwords=10):
        """
        List the top ``Nwords`` words for each topic.
        """
        return [(k, self.list_topic(k, Nwords)) for k in xrange(len(self.phi))]


    def print_topics(self, Nwords=10):
        """
        Print the top ``Nwords`` words for each topic.
        """
        print('Topic\tTop %i words' % 10)
        for k, words in self.list_topics(Nwords):
            print(unicode(k).ljust(3) + '\t' + ' '.join(list(zip(*words))[0]))


    def topic_over_time(self, k, mode='counts', slice_kwargs={}):
        """
        Calculate the representation of topic ``k`` in the corpus over time.
        """

        return self.corpus.feature_distribution('topics', k, mode=mode,
                                                **slice_kwargs)



import os
import iso8601
import logging
import rdflib
import nltk
import codecs
import magic    # To detect file mime-type.
import slate    # PDF processing.
import chardet  # Detect character encodings.
from math import log
logging.basicConfig(level=40)
logging.getLogger('iso8601').setLevel(40)

from datetime import datetime

from tethne import Paper, Corpus, StructuredFeature, StructuredFeatureSet
from tethne.readers.base import RDFParser
from tethne.utilities import _strip_punctuation, mean

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


# RDF terms.
RDF = u'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
DC = u'http://purl.org/dc/elements/1.1/'
FOAF = u'http://xmlns.com/foaf/0.1/'
PRISM = u'http://prismstandard.org/namespaces/1.2/basic/'
RSS = u'http://purl.org/rss/1.0/modules/link/'

URI_ELEM = rdflib.URIRef("http://purl.org/dc/terms/URI")
TYPE_ELEM = rdflib.term.URIRef(RDF + u'type')
VALUE_ELEM = rdflib.URIRef(RDF + u'value')
LINK_ELEM = rdflib.URIRef(RSS + u"link")
FORENAME_ELEM = rdflib.URIRef(FOAF + u'givenname')
SURNAME_ELEM = rdflib.URIRef(FOAF + u'surname')
VOL = rdflib.term.URIRef(PRISM + u'volume')
IDENT = rdflib.URIRef(DC + u"identifier")
TITLE = rdflib.term.URIRef(DC + u'title')


# Build a cost dictionary, assuming Zipf's law and cost = -math.log(probability).
from tethne.readers._rankedwords import WORDS
WORDCOST = dict((k, log((i+1)*log(len(WORDS)))) for i, k in enumerate(WORDS))
MAXWORD = max(len(x) for x in WORDS)

def _infer_spaces(s):
    """
    Uses dynamic programming to infer the location of spaces in a string
    without spaces.
    """
    s = s.lower()

    # Find the best match for the i first characters, assuming cost has
    # been built for the i-1 first characters.
    # Returns a pair (match_cost, match_length).
    def best_match(i):
        candidates = enumerate(reversed(cost[max(0, i - MAXWORD):i]))
        return min((c + WORDCOST.get(s[i-k-1: i], 9e999), k + 1)
                    for k, c in candidates)

    # Build the cost array.
    cost = [0]
    for i in range(1, len(s) + 1):
        c, k = best_match(i)
        cost.append(c)

    # Backtrack to recover the minimal-cost string.
    out = []
    i = len(s)
    while i > 0:
        c, k = best_match(i)
        assert c == cost[i]
        out.append(s[i-k:i])
        i -= k

    return u" ".join(reversed(out))


def extract_text(fpath):
    """
    Extracts structured text content from a plain-text file at ``fpath``.

    Parameters
    ----------
    fpath : str
        Path to the text file..

    Returns
    -------
    :class:`.StructuredFeature`
        A :class:`.StructuredFeature` that contains sentence context.
    """
    with codecs.open(fpath, 'r') as f:  # Determine the encoding of the file.
        document = f.read()
    encoding = chardet.detect(document)['encoding']
    document = document.decode(encoding)

    tokens = []
    sentences = []

    i = 0
    for sentence in nltk.tokenize.sent_tokenize(document):
        sentences.append(i)

        for word in nltk.tokenize.word_tokenize(sentence):
            tokens.append(word)
            i += 1

    contexts = [('page', pages), ('sentence', sentences)]
    return StructuredFeature(tokens, contexts)


def extract_pdf(fpath):
    """
    Extracts structured text content from a PDF at ``fpath``.

    Parameters
    ----------
    fpath : str
        Path to the PDF.

    Returns
    -------
    :class:`.StructuredFeature`
        A :class:`.StructuredFeature` that contains page and sentence contexts.
    """

    with codecs.open(fpath, 'r') as f:  # Determine the encoding of the file.
        document = slate.PDF(f)
    encoding = chardet.detect(document[0])

    tokens = []
    pages = []
    sentences = []

    tokenizer = nltk.tokenize.TextTilingTokenizer()

    i = 0
    for page in document:
        pages.append(i)

        # Decode using the correct encoding.
        page = page.decode(encoding['encoding'])
        for sentence in nltk.tokenize.sent_tokenize(page):
            sentences.append(i)

            for word in nltk.tokenize.word_tokenize(sentence):
                if len(word) > 15:
                    words = nltk.tokenize.word_tokenize(_infer_spaces(word))
                    if mean([len(w) for w in words]) > 2:
                        for w in words:
                            tokens.append(w)
                            i += 1
                        continue

                tokens.append(word)
                i += 1

    contexts = [('page', pages), ('sentence', sentences)]
    return StructuredFeature(tokens, contexts)


class ZoteroParser(RDFParser):
    """
    Reads Zotero RDF files.
    """

    entry_class = Paper
    entry_elements = ['bib:Illustration', 'bib:Recording', 'bib:Legislation',
                      'bib:Document', 'bib:BookSection', 'bib:Book', 'bib:Data',
                      'bib:Letter', 'bib:Report', 'bib:Article',
                      'bib:Manuscript', 'bib:Image',
                      'bib:ConferenceProceedings', 'bib:Thesis']
    tags = {
        'isPartOf': 'journal'
    }

    meta_elements = [
        ('date', rdflib.URIRef("http://purl.org/dc/elements/1.1/date")),
        ('identifier',
         rdflib.URIRef("http://purl.org/dc/elements/1.1/identifier")),
        ('abstract', rdflib.URIRef("http://purl.org/dc/terms/abstract")),
        ('authors_full', rdflib.URIRef("http://purl.org/net/biblio#authors")),
        ('link', rdflib.URIRef("http://purl.org/rss/1.0/modules/link/link")),
        ('title', rdflib.URIRef("http://purl.org/dc/elements/1.1/title")),
        ('isPartOf', rdflib.URIRef("http://purl.org/dc/terms/isPartOf")),
        ('pages', rdflib.URIRef("http://purl.org/net/biblio#pages")),
        ('documentType',
         rdflib.URIRef("http://www.zotero.org/namespaces/export#itemType"))]

    def __init__(self, path, **kwargs):
        name = os.path.split(path)[1]
        path = os.path.join(path, '{0}.rdf'.format(name))
        super(ZoteroParser, self).__init__(path, **kwargs)

        self.full_text = {}     # Collect StructuredFeatures until finished.

    def open(self):
        """
        Fixes RDF validation issues. Zotero incorrectly uses ``rdf:resource`` as
        a child element for Attribute; ``rdf:resource`` should instead be used
        as an attribute of ``link:link``.
        """

        with open(self.path, 'r') as f:
            corrected = f.read().replace('rdf:resource rdf:resource',
                                         'link:link rdf:resource')
        with open(self.path, 'w') as f:
            f.write(corrected)

        super(ZoteroParser, self).open()

    def handle_identifier(self, value):
        """

        """

        identifier = unicode(self.graph.value(subject=value, predicate=VALUE_ELEM))
        ident_type = self.graph.value(subject=value, predicate=TYPE_ELEM)
        if ident_type == URI_ELEM:
            self.set_value('uri', identifier)


    def handle_link(self, value):
        """
        rdf:link rdf:resource points to the resource described by a record.
        """
        for s, p, o in self.graph.triples((value, None, None)):
            if p == LINK_ELEM:
                return unicode(o).replace('file://', '')

    def handle_date(self, value):
        """
        Attempt to coerced date to ISO8601.
        """
        try:
            return iso8601.parse_date(unicode(value)).year
        except iso8601.ParseError:
            for datefmt in ("%B %d, %Y", "%Y-%m", "%Y-%m-%d", "%m/%d/%Y"):
                try:
                    # TODO: remove str coercion.
                    return datetime.strptime(unicode(value), datefmt).date().year
                except ValueError:
                    pass

    def handle_documentType(self, value):
        return value

    def handle_authors_full(self, value):
        authors = [self.handle_author(o) for s, p, o
                   in self.graph.triples((value, None, None))]
        return [a for a in authors if a is not None]

    def handle_author(self, value):
        forename_iter = self.graph.triples((value, FORENAME_ELEM, None))
        surname_iter = self.graph.triples((value, SURNAME_ELEM, None))
        norm = lambda s: unicode(s).upper().replace('.', '')

        # TODO: DRY this out.
        try:
            forename = norm([e[2] for e in forename_iter][0])
        except IndexError:
            forename = ''

        try:
            surname = norm([e[2] for e in surname_iter][0])
        except IndexError:
            surname = ''

        if surname == '' and forename == '':
            return
        return surname, forename

    def handle_isPartOf(self, value):
        journal = None
        for s, p, o in self.graph.triples((value, None, None)):

            if p == VOL:        # Volume number
                self.set_value('volume', unicode(o))
            elif p == TITLE:
                journal = unicode(o)    # Journal title.
        return journal

    def handle_pages(self, value):
        return tuple(value.split('-'))

    def postprocess_pages(self, entry):
        if len(entry.pages) == 1:
            start = entry.pages
            end = None
        else:
            start, end = entry.pages
        setattr(entry, 'pageStart', start)
        setattr(entry, 'pageEnd', end)
        del entry.pages

    def postprocess_link(self, entry):
        """
        Attempt to load full-text content from resource.
        """

        if not self.follow_links:
            return

        if type(entry.link) is not list:
            entry.link = [entry.link]

        for link in list(entry.link):
            if not os.path.exists(link):
                continue

            mime_type = magic.from_file(link, mime=True)
            if mime_type == 'application/pdf':
                structuredfeature = extract_pdf(link)
            elif mime_type == 'text/plain':
                structuredfeature = extract_text(link)
            else:
                structuredfeature = None

            if not structuredfeature:
                continue

            fset_name = mime_type.split('/')[-1] + '_text'
            if not fset_name in self.full_text:
                self.full_text[fset_name] = {}

            if hasattr(self, 'index_by'):
                ident = getattr(entry, self.index_by)
            else:   # If `index_by` is not set, use `uri` by default.
                ident = entry.uri

            self.full_text[fset_name][ident] = structuredfeature


def read(path, corpus=True, index_by='uri', follow_links=True, **kwargs):
    """
    Read bibliographic data from Zotero RDF.

    Example
    -------
    Assuming that the Zotero collection was exported to the directory
    ``/my/working/dir`` with the name ``myCollection``, a subdirectory should
    have been created at ``/my/working/dir/myCollection``, and an RDF file
    should exist at ``/my/working/dir/myCollection/myCollection.rdf``.

    .. code-block:: python

       >>> from tethne.readers.zotero import read
       >>> myCorpus = read('/my/working/dir/myCollection')
       >>> myCorpus
       <tethne.classes.corpus.Corpus object at 0x10047e350>


    Parameters
    ----------
    path : str
        Path to the output directory created by Zotero. Expected to contain a
        file called ``[directory_name].rdf``.
    corpus : bool
        (default: True) If True, returns a :class:`.Corpus`\. Otherwise,
        returns a list of :class:`.Paper`\s.
    index_by : str
        (default: ``'identifier'``) :class:`.Paper` attribute name to use as
        the primary indexing field. If the field is missing on a
        :class:`.Paper`\, a unique identifier will be generated based on the
        title and author names.
    follow_links : bool
        If ``True``, attempts to load full-text content from attached files
        (e.g. PDFs with embedded text).
    kwargs : kwargs
        Passed to the :class:`.Corpus` constructor.

    Returns
    -------
    corpus : :class:`.Corpus`
    """
    # TODO: is there a case where `from_dir` would make sense?

    parser = ZoteroParser(path, index_by=index_by, follow_links=follow_links)
    papers = parser.parse()

    if corpus:
        c = Corpus(papers, index_by=index_by, **kwargs)
        for fset_name, fset_values in parser.full_text.items():
            c.features[fset_name] = StructuredFeatureSet(fset_values)
        return c
    return papers



"""
Classes in this module provide structures for additional data about
:class:`.Paper`\s.
"""

from collections import Counter, defaultdict

from tethne.utilities import _iterable
try:    # Might as well use numpy if it is available.
    import numpy as np
    argsort = lambda l: list(np.argsort(l))
except ImportError:
    from tethne.utilities import argsort

import logging
logger = logging.getLogger('feature')
logger.setLevel('WARNING')


from itertools import chain
from collections import Counter, defaultdict

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    xrange = range
    unicode = str


class StructuredFeature(list):
    """
    A :class:`.StructuredFeature` represents the contents of a document as an
    array of tokens, divisible into a set of nested contexts.

    The canonical use-case is to represent a document as a set of words divided
    into sentences, paragraphs, and (perhaps) pages.

    Parameters
    ----------
    tokens : list
        An ordered list of tokens.
    contexts : list
        A list of (name, indices) 2-tuples, where ``name`` is string-like and
        indices is an iterable of int token indices.
    reference : tuple
        A (feature, map) 2-tuple, where ``feature`` is a
        :class:`.StructuredFeature` and ``map`` is a dict mapping token indices
        in this :class:`.StructuredFeature` to token indices in ``feature``.
    """
    def __init__(self, tokens, contexts=None, reference=None):
        self.extend(tokens)
        self.contexts = {}
        self.contexts_ranked = []
        self.referenceFeature = None
        self.referenceMap = None

        if contexts:
            self._validate_contexts(contexts)
            for context in contexts:
                self._validate_context(context)
                self.add_context(*context)

        if reference:
            self._validate_reference(reference)

            self.referenceFeature, self.referenceMap = reference

    @property
    def unique(self):
        """
        The `set` of unique elements in this :class:`.Feature`\.
        """
        if len(self) > 0:
            return set(self)
        return set()

    def __getitem__(self, selector):
        if type(selector) is int:
            return super(StructuredFeature, self).__getitem__(selector)

        if type(selector) in [str, unicode]:
            if selector in self.contexts:
                return self.context_chunks(selector)
        elif type(selector) is tuple:
            if selector[0] in self.contexts:
                return self.context_chunk(*selector)

    def context_chunks(self, context):
        """
        Retrieves all tokens, divided into the chunks in context ``context``.

        Parameters
        ----------
        context : str
            Context name.

        Returns
        -------
        chunks : list
            Each item in ``chunks`` is a list of tokens.
        """
        N_chunks = len(self.contexts[context])
        chunks = []
        for j in xrange(N_chunks):
            chunks.append(self.context_chunk(context, j))
        return chunks

    def context_chunk(self, context, j):
        """
        Retrieve the tokens in the ``j``th chunk of context ``context``.

        Parameters
        ----------
        context : str
            Context name.
        j : int
            Index of a context chunk.

        Returns
        -------
        chunk : list
            List of tokens in the selected chunk.
        """

        N_chunks = len(self.contexts[context])
        start = self.contexts[context][j]
        if j == N_chunks - 1:
            end = len(self)
        else:
            end = self.contexts[context][j+1]
        return [self[i] for i in xrange(start, end)]

    def _validate_context(self, context):
        try:
            assert hasattr(context, '__iter__')
            assert len(context) == 2
            assert type(context[0]) in [str, unicode]
            assert hasattr(context[1], '__iter__')
            assert type(context[1][0]) is int
        except AssertionError:
            raise ValueError("""a context should be a (name, indices) 2-tuple,
            where ``name`` is string-like and indices is an iterable of int
            token indices.""")

        if max(context[1]) > len(self):
            raise ValueError("""One or more indices in the specified context
            exceed the number of tokens in this StructuredFeature.""")

    @staticmethod
    def _validate_contexts(contexts):
        try:
            assert hasattr(contexts, '__iter__')
        except AssertionError:
            raise ValueError("""contexts should be a list of (name, indices)
            2-tuples, where ``name`` is string-like and indices is an iterable
            of int token indices.""")

    @staticmethod
    def _validate_reference(reference):
        try:
            assert type(reference) is tuple
            assert type(reference[0]) is StructuredFeature
            assert type(reference[1]) is dict
        except AssertionError:
            raise ValueError("""reference should be a (feature, map) 2-tuple
            where ``feature`` is a StructuredFeature and ``map`` is a dict
            mapping token indices in this StructuredFeature to token indices
            in ``feature``.""")

    def add_context(self, name, indices, level=None):
        """
        Add a new context level to the hierarchy.

        By default, new contexts are added to the lowest level of the hierarchy.
        To insert the context elsewhere in the hierarchy, use the ``level``
        argument. For example, ``level=0`` would insert the context at the
        highest level of the hierarchy.

        Parameters
        ----------
        name : str
        indices : list
            Token indices at which each chunk in the context begins.
        level : int
            Level in the hierarchy at which to insert the context. By default,
            inserts context at the lowest level of the hierarchy

        """

        self._validate_context((name, indices))

        if level is None:
            level = len(self.contexts_ranked)
        self.contexts_ranked.insert(level, name)
        self.contexts[name] = indices


class Feature(list):
    """
    A :class:`.Feature` instance is a sparse vector of features over a given
    concept (usually a :class:`.Paper`\).

    For example, a :class:`.Feature` might represent word counts for a single
    :class:`.Paper`\.

    A :class:`.Feature` may be initialized from a list of ``(feature, value)``
    tuples...

    .. code-block:: python

       >>> myFeature = Feature([('the', 2), ('pine', 1), ('trapezoid', 5)])

    ...or by passing a list of raw feature tokens:

    .. codeblock:: python

       >>> myFeature = Feature(['the', 'the', 'pine', 'trapezoid', 'trapezoid',
       ...                      'trapezoid', 'trapezoid', 'trapezoid'])
       >>> myFeature
       [('the', 2), ('pine', 1), ('trapezoid', 5)]

    To get the set of unique features in this :class:`.Feature`\, use
    :prop:`.Feature.unique`\:

    .. code-block:: python

       >>> myFeature.unique
       set(['the', 'pine', 'trapezoid'])

    Normalized feature values (so that all values sum to 1.) can be accessed
    using :prop:`.Feature.norm`\.

    .. code-block:: python

       >>> myFeature.norm
       [('the', 0.25), ('pine', 0.125), ('trapezoid', 0.625)]

    """

    def __init__(self, data):
        if len(data) > 0:
            self.extend(data)

    def __add__(self, data):
        if len(data) > 0:
            if type(data[0]) is tuple and type(data[0][-1]) in [float, int]:
                # There may be overlap with existing features,
                combined_data = defaultdict(type(data[0][-1]))
                for k, v in data + list(self):
                    combined_data[k] += v
                return combined_data.items()
            else:   # Recurses.
                c = Counter(_iterable(data))
                keys = list(c.keys())
                return self.__add__(list(zip(keys, c.values())))
        return self

    def __sub__(self, data):
        if len(data) > 0:
            if type(list(data)[0]) is tuple and type(list(data)[0][-1]) in [float, int]:
                combined_data = defaultdict(type(data[0][-1]))
                combined_data.update(dict(self))
                for k, v in data:
                    combined_data[k] -= v
                return list(combined_data.items())
            else:   # Recurses.
                return self.__sub__(list(Counter(_iterable(data)).items()))
        return self

    def __iadd__(self, data):
        return self.extend(data)

    def __isub__(self, data):
        if len(data) > 0:
            combined_data = self.__sub__(data)
            del self[:]
            super(Feature, self).extend(combined_data)
        return self

    def extend(self, data):
        if len(data) > 0:
            combined_data = self.__add__(data)  # Combines new and extant data.
            del self[:]                         # Clear old data.
            super(Feature, self).extend(combined_data)
        return self

    @property
    def unique(self):
        """
        The `set` of unique elements in this :class:`.Feature`\.
        """
        if len(self) > 0:
            return set(list(zip(*self))[0])
        return set()

    @property
    def norm(self):
        T = sum(list(zip(*self))[1])
        return Feature([(i, float(v)/T) for i, v in self])

    def top(self, topn=10):
        """
        Get a list of the top ``topn`` features in this :class:`.Feature`\.

        Example
        -------

        .. code-block:: python

        >>> myFeature = Feature([('the', 2), ('pine', 1), ('trapezoid', 5)])
        >>> myFeature.top(1)
        [('trapezoid', 5)]

        Parameters
        ----------
        topn : int

        Returns
        -------
        list
        """
        return [self[i] for i in argsort(list(zip(*self))[1])[::-1][:topn]]

    def value(self, element):

        return dict(self)[element]


class BaseFeatureSet(object):
    def __init__(self, features={}):
        self._setUp()

        for paper, feature in features.items():
            self.add(paper, feature)

    def _setUp(self):
        self.index = {}
        self.lookup = {}
        self.counts = {}
        self.documentCounts = {}
        self.features = {}
        self.with_feature = {}

    def __getitem__(self, key):
        try:
            return self.features[key]
        except KeyError as E:
            if type(key) is int:
                return self.features.values()[key]
            raise E

    def __len__(self):
        return len(self.features)

    def items(self):
        return self.features.items()

    @property
    def unique(self):
        """
        The `set` of unique elements in this :class:`.FeatureSet`\.
        """
        return set(self.lookup.keys())

    @property
    def N_features(self):
        return len(self.unique)

    @property
    def N_documents(self):
        return len(self.features)

    def count(self, elem):
        logger.debug(u'Get count for {0}'.format(elem))
        if elem in self.lookup:
            i = self.lookup[elem]
            count = self.counts[i]
            logger.debug(u'Found elem %s with index %i and count %f' % (elem, i, count))
            return count
        else:
            return 0.

    def documentCount(self, elem):
        if elem in self.lookup:
            return self.documentCounts[self.lookup[elem]]
        else:
            return 0.

    def papers_containing(self, elem):
        return self.with_feature[self.lookup[elem]]

    def add(self, paper_id, feature):
        if type(feature) not in [Feature, StructuredFeature]:
            raise ValueError("""`feature` must be an instance of Feature or
            StructuredFeature""")

        self.features[paper_id] = feature

        self.extend_index(feature.unique)
        for item in feature:
            if type(item) is tuple:
                elem, count = item
            else:
                elem = item
                count = 1.

            self.counts[self.lookup[elem]] += count
            self.documentCounts[self.lookup[elem]] += 1.
            self.with_feature[self.lookup[elem]].append(paper_id)

    def extend_index(self, elements):
        """

        """
        new_elements = list(set(elements) - self.unique)
        for i in xrange(len(new_elements)):
            x = len(self.index)
            self.index[x] = new_elements[i]
            self.lookup[new_elements[i]] = x
            self.counts[x] = 0.
            self.documentCounts[x] = 0.
            self.with_feature[x] = []

    def top(self, topn, by='counts'):
        """
        Get the top ``topn`` features in the :class:`.FeatureSet`\.

        Parameters
        ----------
        topn : int
            Number of features to return.
        by : str
            (default: 'counts') How features should be sorted. Must be 'counts' or
            'documentcounts'.

        Returns
        -------
        list

        """

        if by not in ['counts', 'documentCounts']:
            raise NameError('kwarg `by` must be "counts" or "documentCounts"')

        cvalues = getattr(self, by)
        top = [list(cvalues.keys())[i] for i in argsort(list(cvalues.values()))][::-1]
        return [(self.index[x], cvalues[x]) for x in top][:topn]


class StructuredFeatureSet(BaseFeatureSet):
    """
    A :class:`.StructuredFeatureSet` organizes several
    :class:`.StructuredFeature` instances.
    """

    def transform(self, func):
        features = {}
        for i, feature in self.features.items():
            feature_ = []
            for f in feature:
                t = self.lookup[f]
                v_ = func(f, self.count(f), feature.count(f), self.documentCounts[t])
                if v_ is not None:
                    feature_.append(f)
            features[i] = StructuredFeature(feature_)

        return StructuredFeatureSet(features)

    def context_chunks(self, context):
        """
        Retrieves all tokens, divided into the chunks in context ``context``.

        If ``context`` is not found in a feature, then the feature will be
        treated as a single chunk.

        Parameters
        ----------
        context : str
            Context name.

        Returns
        -------
        papers : list
            2-tuples of (paper ID, chunk indices).
        chunks : list
            Each item in ``chunks`` is a list of tokens.
        """

        chunks = []
        papers = []
        for paper, feature in self.features.items():
            if context in feature.contexts:
                new_chunks = feature.context_chunks(context)
            else:
                new_chunks = list(feature)
            indices = range(len(chunks), len(chunks) + len(new_chunks))
            papers.append((paper, indices))
            chunks += new_chunks
        return papers, chunks


class FeatureSet(BaseFeatureSet):
    """
    A :class:`.FeatureSet` organizes multiple :class:`.Feature` instances.
    """

    def __init__(self, features=None):
        if not features:
            features = dict()
        self._setUp()

        logger.debug(u'Initialize FeatureSet with %i features' % len(features))
        self.features = features
        allfeatures = [v for v in chain(*features.values())]
        logger.debug('features: {0}; allfeatures: {1}'.format(len(features), len(allfeatures)))
        if len(features) > 0 and len(allfeatures) > 0:
            allfeatures_keys = zip(*allfeatures)[0]

            for i, elem in enumerate(set(allfeatures_keys)):
                self.index[i] = elem
                self.lookup[elem] = i
                logger.debug(u'Add feature {0} with index {1}'.format(elem, i))

            self.counts = defaultdict(float)
            for elem, v in allfeatures:
                i = self.lookup[elem]
                self.counts[i] += v
            self.documentCounts = Counter([self.lookup[elem]
                                           for elem
                                           in allfeatures_keys])

            self.with_feature = defaultdict(list)
            for paper_id, counts in features.items():
                try:
                    for elem in zip(*counts)[0]:
                        i = self.lookup[elem]
                        self.with_feature[i].append(paper_id)
                except IndexError:    # A Paper may not have any features.
                    pass


    def transform(self, func):
        """
        Apply a transformation to tokens in this :class:`.FeatureSet`\.

        Parameters
        ----------
        func : callable
            Should take four parameters: token, value in document (e.g. count),
            value in :class:`.FeatureSet` (e.g. overall count), and document
            count (i.e. number of documents in which the token occurs). Should
            return a new numeric (int or float) value, or None. If value is 0
            or None, the token will be excluded.

        Returns
        -------
        :class:`.FeatureSet`

        Examples
        --------

        Apply a tf*idf transformation.

        .. code-block:: python

           >>> words = corpus.features['words']
           >>> def tfidf(f, c, C, DC):
           ... tf = float(c)
           ... idf = log(float(len(words.features))/float(DC))
           ... return tf*idf
           >>> corpus.features['words_tfidf'] = words.transform(tfidf)

        """
        features = {}
        for i, feature in self.features.items():
            feature_ = []
            for f, v in feature:
                t = self.lookup[f]
                v_ = func(f, v, self.counts[t], self.documentCounts[t])
                if v_ > 0 and v_ is not None:
                    feature_.append((f, v_))
            features[i] = Feature(feature_)

        return FeatureSet(features)

    def translate(self, func):
        features = {}
        for i, feature in self.features.items():
            features_ = []
            for f, v in feature:
                t = self.lookup[f]
                f_ = func(f, v, self.counts[t], self.documentCounts[t])
                if f_:
                    feature_.append((f_, v))
            features[i] = Feature(feature_)
        return FeatureSet(features)


    def as_matrix(self):
        """

        """

        matrix = [[0. for e in xrange(self.N_features)]
                  for i in xrange(self.N_documents)]
        for i, p in enumerate(self.features.keys()):
            f = self.features[p]
            for e, c in f:
                j = self.lookup[e]
                matrix[i][j] = c

        return matrix

    def as_vector(self, p, norm=False):
        m = len(self.index.keys())

        if norm:
            values = dict(self.features[p].norm)
        else:
            values = dict(self.features[p])

        vect = []
        for i in xrange(m):
            e = self.index[i]
            if e in values:
                c = float(values[e])
            else:
                c = 0.
            vect.append(c)

        return vect


def feature(f):
    """
    Decorator for properties that should be represented as :class:`.Feature`\s.
    """
    def deco(self):
        return Feature(f(self))
    return deco



"""
Parser for Web of Science field-tagged bibliographic data.

Tethne parsers Web of Science field-tagged data into a set of
:class:`.Paper`\s, which are then encapsulated in a :class:`.Corpus`\. The
:class:`.WoSParser` can be instantiated directly, or you can simply use
:func:`.read` to parse a single file or a directory containing several data
files.

.. code-block:: python

   >>> from tethne.readers import wos
   >>> corpus = wos.read("/path/to/some/wos/data")
   >>> corpus
   <tethne.classes.corpus.Corpus object at 0x10057c2d0>

"""

import re
import os

from tethne.readers.base import FTParser
from tethne import Corpus, Paper
from tethne.utilities import _strip_punctuation, _space_sep, strip_tags, is_number

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


class WoSParser(FTParser):
    """
    Parser for Web of Science field-tagged data.

    .. code-block:: python

       >>> from tethne.readers.wos import WoSParser
       >>> parser = WoSParser("/path/to/download.txt")
       >>> papers = parser.read()

    """

    start_tag = 'PT'
    """
    Field-tag used to mark the start of a record.
    """

    end_tag = 'ER'
    """
    Field-tag used to mark the end of a record.
    """

    concat_fields = ['abstract', 'keywords', 'funding', 'title', 'references',
                     'journal']
    """
    Fields that span multiple lines that should be concatenated into a single
    value.
    """

    entry_class = Paper
    """
    The class that should be used to represent a single bibliographic record.
    This can be changed to support more sophisticated data models.
    """

    tags = {
        'PY': 'date',
        'SO': 'journal',
        'AB': 'abstract',
        'TI': 'title',
        'AF': 'authors_full',
        'AU': 'authors_init',
        'ID': 'authorKeywords',
        'DE': 'keywordsPlus',
        'DI': 'doi',
        'BP': 'pageStart',
        'EP': 'pageEnd',
        'VL': 'volume',
        'IS': 'issue',
        'CR': 'citedReferences',
        'DT': 'documentType',
        'CA': 'groupAuthors',
        'ED': 'editors',
        'SE': 'bookSeriesTitle',
        'BS': 'bookSeriesSubtitle',
        'LA': 'language',
        'DT': 'documentType',
        'CT': 'conferenceTitle',
        'CY': 'conferenceDate',
        'HO': 'conferenceHost',
        'CL': 'conferenceLocation',
        'SP': 'conferenceSponsors',
        'C1': 'authorAddress',
        'RP': 'reprintAddress',
        'EM': 'emailAddress',
        'FU': 'funding',
        'NR': 'citationCount',
        'TC': 'timesCited',
        'PU': 'publisher',
        'PI': 'publisherCity',
        'PA': 'publisherAddress',
        'SC': 'subject',
        'SN': 'ISSN',
        'BN': 'ISSN',
        'UT': 'wosid',
        'JI': 'isoSource',
    }
    """
    Maps field-tags onto field names.
    """

    def parse_author(self, value):
        """
        Attempts to split an author name into last and first parts.
        """
        tokens = tuple([t.upper().strip() for t in value.split(',')])
        if len(tokens) == 1:
            tokens = value.split(' ')
        if len(tokens) > 0:
            if len(tokens) > 1:
                aulast, auinit = tokens[0:2]    # Ignore JR, II, III, etc.
            else:
                aulast = tokens[0]
                auinit = ''
        else:
            aulast, auinit = tokens[0], ''
        aulast = _strip_punctuation(aulast).upper()
        auinit = _strip_punctuation(auinit).upper()
        return aulast, auinit

    def handle_AF(self, value):
        return self.parse_author(value)

    def handle_PY(self, value):
        """
        WoS publication years are cast to integers.
        """
        return int(value)

    def handle_AU(self, value):
        aulast, auinit = self.parse_author(value)
        auinit = _space_sep(auinit)   # Separate author initials with spaces.
        return aulast, auinit

    def handle_TI(self, value):
        """
        Convert article titles to Title Case.
        """
        return unicode(value).title()

    def handle_VL(self, value):
        """
        Volume should be a unicode string, even if it looks like an integer.
        """
        return unicode(value)

    def handle_CR(self, value):
        """
        Parses cited references.
        """
        citation = self.entry_class()

        value = strip_tags(value)

        # First-author name and publication date.
        ptn = '([\w\s\W]+),\s([0-9]{4}),\s([\w\s]+)'
        ny_match = re.match(ptn, value, flags=re.U)
        nj_match = re.match('([\w\s\W]+),\s([\w\s]+)',
                            value, flags=re.U)
        if ny_match is not None:
            name_raw, date, journal = ny_match.groups()
        elif nj_match is not None:
            name_raw, journal = nj_match.groups()
            date = None
        else:
            return

        datematch = re.match('([0-9]{4})', value)
        if datematch:
            date = datematch.group(1)
            name_raw = None

        if name_raw:
            name_tokens = [t.replace('.', '') for t in name_raw.split(' ')]
            if len(name_tokens) > 4 or value.startswith('*'):    # Probably not a person.
                proc = lambda x: _strip_punctuation(x)
                aulast = ' '.join([proc(n) for n in name_tokens]).upper()
                auinit = ''
            elif len(name_tokens) > 0:
                aulast = name_tokens[0].upper()
                proc = lambda x: _space_sep(_strip_punctuation(x))
                auinit = ' '.join([proc(n) for n in name_tokens[1:]]).upper()
            else:
                aulast = name_tokens[0].upper()
                auinit = ''

            setattr(citation, 'authors_init', [(aulast, auinit)])

        if date:
            date = int(date)
        setattr(citation, 'date', date)
        setattr(citation, 'journal', journal)

        # Volume.
        v_match = re.search('\,\s+V([0-9A-Za-z]+)', value)
        if v_match is not None:
            volume = v_match.group(1)
        else:
            volume = None
        setattr(citation, 'volume', volume)

        # Start page.
        p_match = re.search('\,\s+[Pp]([0-9A-Za-z]+)', value)
        if p_match is not None:
            page = p_match.group(1)
        else:
            page = None
        setattr(citation, 'pageStart', page)

        # DOI.
        doi_match = re.search('DOI\s(.*)', value)
        if doi_match is not None:
            doi = doi_match.group(1)
        else:
            doi = None
        setattr(citation, 'doi', doi)
        return citation

    def postprocess_authorKeywords(self, entry):
        """
        Parse author keywords.

        Author keywords are usually semicolon-delimited.
        """

        if type(entry.authorKeywords) not in [str, unicode]:
            aK = u' '.join([unicode(k) for k in entry.authorKeywords])
        else:
            aK = entry.authorKeywords
        entry.authorKeywords = [k.strip().upper() for k in aK.split(';')]

    def postprocess_keywordsPlus(self, entry):
        """
        Parse WoS "Keyword Plus" keywords.

        Keyword Plus keywords are usually semicolon-delimited.
        """

        if type(entry.keywordsPlus) in [str, unicode]:
            entry.keywordsPlus = [k.strip().upper() for k
                                  in entry.keywordsPlus.split(';')]

    def postprocess_funding(self, entry):
        """
        Separates funding agency from grant numbers.
        """

        if type(entry.funding) not in [str, unicode]:
            return

        sources = [fu.strip() for fu in entry.funding.split(';')]
        sources_processed = []
        for source in sources:
            m = re.search('(.*)?\s+\[(.+)\]', source)
            if m:
                agency, grant = m.groups()
            else:
                agency, grant = source, None
            sources_processed.append((agency, grant))
        entry.funding = sources_processed

    def postprocess_authors_full(self, entry):
        """
        If only a single author was found, ensure that ``authors_full`` is
        nonetheless a list.
        """
        if type(entry.authors_full) is not list:
            entry.authors_full = [entry.authors_full]

    def postprocess_authors_init(self, entry):
        """
        If only a single author was found, ensure that ``authors_init`` is
        nonetheless a list.
        """
        if type(entry.authors_init) is not list:
            entry.authors_init = [entry.authors_init]

    def postprocess_citedReferences(self, entry):
        """
        If only a single cited reference was found, ensure that
        ``citedReferences`` is nonetheless a list.
        """
        if type(entry.citedReferences) is not list:
            entry.citedReferences = [entry.citedReferences]


def from_dir(path, corpus=True, **kwargs):
    raise DeprecationWarning("from_dir() is deprecated. Use read() instead.")
    papers = []
    for sname in os.listdir(path):
        if sname.endswith('txt') and not sname.startswith('.'):
            papers += read(os.path.join(path, sname), corpus=False)
    if corpus:
        return Corpus(papers, **kwargs)
    return papers


def corpus_from_dir(path, **kwargs):
    raise DeprecationWarning("corpus_from_dir is deprecated in v0.8, use" +
                             " read directly, instead.")
    return read(path, corpus=True, **kwargs)


def read_corpus(path, **kwargs):
    """
    .. DANGER::
       read_corpus is deprecated in v0.8, use :func:`.read` instead.
    """
    raise DeprecationWarning("read_corpus is deprecated in v0.8, use" +
                             " read directly, instead.")
    return read(path, corpus=True, **kwargs)


def read(path, corpus=True, index_by='wosid', **kwargs):
    """
    Parse one or more WoS field-tagged data files.

    Example
    -------
    .. code-block:: python

       >>> from tethne.readers import wos
       >>> corpus = wos.read("/path/to/some/wos/data")
       >>> corpus
       <tethne.classes.corpus.Corpus object at 0x10057c2d0>

    Parameters
    ----------
    path : str
        Path to WoS field-tagged data. Can be a path directly to a single data
        file, or to a directory containing several data files.
    corpus : bool
        If True (default), returns a :class:`.Corpus`\. If False, will return
        only a list of :class:`.Paper`\s.

    Returns
    -------
    :class:`.Corpus` or :class:`.Paper`
    """

    if not os.path.exists(path):
        raise ValueError('No such file or directory')

    if os.path.isdir(path):    # Directory containing 1+ WoS data files.
        papers = []
        for sname in os.listdir(path):
            if sname.endswith('txt') and not sname.startswith('.'):
                papers += read(os.path.join(path, sname), corpus=False)
    else:   # A single data file.
        papers = WoSParser(path).parse()

    if corpus:
        return Corpus(papers, index_by=index_by,  **kwargs)
    return papers



"""
Methods for network analysis.

.. autosummary::
   :nosignatures:
   
   global_closeness_centrality
   
"""

import networkx as nx

def global_closeness_centrality(g, node=None, normalize=True):
    """
    Calculates global closeness centrality for one or all nodes in the network.

    See :func:`.node_global_closeness_centrality` for more information.

    Parameters
    ----------
    g : networkx.Graph
    normalize : boolean
        If True, normalizes centrality based on the average shortest path
        length. Default is True.

    Returns
    -------
    C : dict
        Dictionary of results, with node identifiers as keys and gcc as values.
    """

    if not node:
        C = {}
        for node in g.nodes():
            C[node] = global_closeness_centrality(g, node, normalize=normalize)
        return C

    values = nx.shortest_path_length(g, node).values()
    c = sum([1./pl for pl in values if pl != 0.]) / len(g)

    if normalize:
        ac = 0
        for sg in nx.connected_component_subgraphs(g):
            if len(sg.nodes()) > 1:
                aspl = nx.average_shortest_path_length(sg)
                ac += (1./aspl) * (float(len(sg)) / float(len(g))**2 )
        c = c/ac

    return c



"""
"""

import os
import re
import shutil
import shutil
import tempfile
import subprocess


class Model(object):
    """
    Base class for models.
    """

    def __init__(self, corpus, **kwargs):
        """
        Initialize the ModelManager.
        """

        self.corpus = corpus
        self.temp = tempfile.mkdtemp()

        for attr in ['run', 'prep']:
            if not hasattr(self, attr):
                msg = 'Model subclass must implement method {0}'.format(attr)
                raise AttributeError(msg)

        # Make all kwargs available as attributes.
        for key, value in kwargs.items():
            if hasattr(self, key):  # Don't overwrite methods.
                if hasattr(getattr(self, key), '__call__'):
                    continue
            setattr(self, key, value)

        self.prep()

    def __del__(self):
        """
        Delete temporary directory and all files contained therein.
        """
        if hasattr(self, 'nodelete'):
            if self.nodelete:
                return
        shutil.rmtree(self.temp)

    @property
    def ll_trace(self):
        """
        Plots LL/topic over iterations.
        """

        return self.ll

    def fit(self, **kwargs):
        # Make all kwargs available as attributes.
        for key, value in kwargs.items():
            if hasattr(self, key):  # Don't overwrite methods.
                if hasattr(getattr(self, key), '__call__'):
                    continue
            setattr(self, key, value)

        self.run(**kwargs)



"""
Methods for parsing JSTOR Data-for-Research datasets.

.. autosummary::


"""

import os
import xml.etree.ElementTree as ET
import re
from collections import Counter
from tethne import Paper, Corpus, Feature, FeatureSet
from tethne.utilities import dict_from_node, strip_non_ascii, number
from tethne.readers.base import XMLParser
import iso8601

from unidecode import unidecode

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


class DfRParser(XMLParser):
    entry_class = Paper

    tags = {
        'type': 'documentType',
        'pubdate': 'date',
        'journaltitle': 'journal',
        'author': 'authors_full',
    }

    def open(self):
        with open(self.path, 'r') as f:
            # JSTOR hasn't always represented ampersands correctly.
            contents = re.sub('(&)(?!amp;)', lambda match: '&amp;', f.read())
            self.root = ET.fromstring(contents)
        pattern = './/{elem}'.format(elem=self.entry_element)
        self.elements = self.root.findall(pattern)

        self.at_start = False
        self.at_end = False
        self.children = []

    def handle_unicode(self, value):
        if type(value) is not str:
            value = unidecode(value)
        return value

    def handle_journaltitle(self, value):
        return self.handle_unicode(value)

    def handle_title(self, value):
        return self.handle_unicode(value)

    def handle_author(self, value):
        if type(value) is not str:
            value = unidecode(value)

        lname = value.split(' ')

        final = lname[-1].upper()
        if final in ['JR.', 'III']:
            aulast = lname[-2].upper() + " " + final.strip(".")
            auinit = ' '.join(lname[0:-2]).replace('.','').strip().upper()
        else:
            aulast = final
            auinit = ' '.join(lname[0:-1]).replace('.','').strip().upper()

        return aulast, auinit

    def handle_pubdate(self, value):
        return iso8601.parse_date(value).year

    def postprocess_authors_full(self, entry):
        if type(entry.authors_full) is not list:
            entry.authors_full = [entry.authors_full]


class GramGenerator(object):
    """
    Yields N-gram data from on-disk dataset, to make loading big datasets a bit
    more memory-friendly.

    Reusable, in the sense that :func:`.items`\, :func:`.items`\,
    :func:`.keys`\, and :func:`.values` all return new :class:`.GramGenerator`
    instances with the same path. This allows a :class:`.GramGenerator` to
    sneakily pass as an ngrams dict in most practical situations.
    """

    def __init__(self, path, elem, values=False, keys=False, ignore_hash=True):
        """

        Parameters
        ----------
        path : str
            Path to unzipped JSTOR DfR folder containing N-grams (e.g.
            'bigrams').
        elem : str
            Element in DfR dataset containing data of interest. E.g. 'bigrams'.
        values : bool
            If True, :func:`.next` returns only values. Otherwise, returns
            (key,value) tuples.
        """

        if not os.path.exists(path):
            raise ValueError('No such file or directory')

        self.path = path
        self.elem = elem
        if elem.endswith('s'):
            self.elem_xml = elem[:-1]
        else:
            self.elem_xml = elem

        self.ignore_hash = ignore_hash

        self.files = os.listdir(os.path.join(path, elem))
        self.N = len([ d for d in self.files if d.split('.')[-1] == 'XML' ])
        self.i = 0

        self.V = values
        self.K = keys

        if self.V and self.K:
            raise ValueError('values and keys cannot both be true.')

    def __len__(self):
        return self.N

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.i < self.N:
            cur = int(self.i)
            self.i += 1
            return self._get(cur)
        else:
            raise StopIteration()

    def items(self):
        """
        Returns a :class:`GramGenerator` that produces key,value tuples.
        """
        return GramGenerator(self.path, self.elem,
                             ignore_hash=self.ignore_hash)

    def items(self):
        """
        Returns a :class:`GramGenerator` that produces key,value tuples.
        """
        return GramGenerator(self.path, self.elem,
                             ignore_hash=self.ignore_hash)

    def values(self):
        """
        Returns a :class:`GramGenerator` that produces only values.
        """
        return GramGenerator(self.path, self.elem, values=True,
                             ignore_hash=self.ignore_hash)

    def keys(self):
        """
        Returns a :class:`GramGenerator` that produces only keys.
        """
        return GramGenerator(self.path, self.elem, keys=True,
                             ignore_hash=self.ignore_hash)

    def __getitem__(self, key):
        return self._get(key)

    def _get(self, i):
        """
        Retrieve data for the ith file in the dataset.
        """
        with open(os.path.join(self.path, self.elem, self.files[i]), 'r') as f:
            # JSTOR hasn't always produced valid XML.
            contents = re.sub('(&)(?!amp;)', lambda match: '&amp;', f.read())
            root = ET.fromstring(contents)
        doi = root.attrib['id']

        if self.K:  # Keys only.
            return doi

        grams = []
        for gram in root.findall(self.elem_xml):
            text = unidecode(unicode(gram.text.strip()))
            if ( not self.ignore_hash or '#' not in list(text) ):
                c = ( text, number(gram.attrib['weight']) )
                grams.append(c)

        if self.V:  # Values only.
            return grams

        return doi, grams   # Default behavior.

def _get_citation_filename(basepath):
    for fname in ["citations.xml", "citations.XML"]:
        if os.path.exists(os.path.join(basepath, fname)):
            return fname


def read(path, corpus=True, index_by='doi', **kwargs):
    """
    Yields :class:`.Paper` s from JSTOR DfR package.

    Each :class:`.Paper` is tagged with an accession id for this
    read/conversion.

    Parameters
    ----------
    filepath : string
        Filepath to unzipped JSTOR DfR folder containing a citations.xml file.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` objects.

    Examples
    --------

    .. code-block:: python

       >>> from tethne.readers import dfr
       >>> papers = dfr.read("/Path/to/DfR")
    """

    citationfname = _get_citation_filename(path)

    papers = []
    if citationfname:   # Valid DfR dataset.
        parser = DfRParser(os.path.join(path, citationfname))
        papers += parser.parse()

    else:   # Possibly a directory containing several DfR datasets?
        papers = []
        # Search for DfR datasets in subdirectories.
        for dirpath, dirnames, filenames in os.walk(path):
            citationfname = _get_citation_filename(dirpath)
            if citationfname:
                papers += read(dirpath, corpus=False)

    if len(papers) == 0:
        raise ValueError('No DfR datasets found at %s' % path)

    if corpus:
        corpus = Corpus(parser.data, index_by=index_by, **kwargs)

        # Find and read N-gram data.
        for sname in os.listdir(path):
            fpath = os.path.join(path, sname)   # Full path.
            if os.path.isdir(fpath) and not sname.startswith('.'):
                datafiles = [f for f in os.listdir(fpath)
                             if f.lower().endswith('xml')]
                if len(datafiles) > 0:
                    corpus.features[sname] = ngrams(path, sname)
        return corpus
    return papers

def ngrams(path, elem, ignore_hash=True):
    """
    Yields N-grams from a JSTOR DfR dataset.

    Parameters
    ----------
    path : string
        Path to unzipped JSTOR DfR folder containing N-grams.
    elem : string
        Name of subdirectory containing N-grams. (e.g. 'bigrams').
    ignore_hash : bool
        If True, will exclude all N-grams that contain the hash '#' character.

    Returns
    -------
    ngrams : :class:`.FeatureSet`

    """

    grams = GramGenerator(path, elem, ignore_hash=ignore_hash)
    return FeatureSet({k: Feature(f) for k, f in grams})


def tokenize(ngrams, min_tf=2, min_df=2, min_len=3, apply_stoplist=False):
    """
    Builds a vocabulary, and replaces words with vocab indices.

    Parameters
    ----------
    ngrams : dict
        Keys are paper DOIs, values are lists of (Ngram, frequency) tuples.
    apply_stoplist : bool
        If True, will exclude all N-grams that contain words in the NLTK
        stoplist.

    Returns
    -------
    t_ngrams : dict
        Tokenized ngrams, as doi:{i:count}.
    vocab : dict
        Vocabulary as i:term.
    token_tf : :class:`.Counter`
        Term counts for corpus, as i:count.
    """

    vocab = {}
    vocab_ = {}
    word_tf = Counter()
    word_df = Counter()
    token_tf = Counter()
    token_df = Counter()
    t_ngrams = {}

    # Get global word counts, first.
    for grams in ngrams.values():
        for g,c in grams:
            word_tf[g] += c
            word_df[g] += 1

    if apply_stoplist:
        stoplist = stopwords.words()

    # Now tokenize.
    for doi, grams in ngrams.items():
        t_ngrams[doi] = []
        for g,c in grams:
            ignore = False

            # Ignore extremely rare words (probably garbage).
            if word_tf[g] < min_tf or word_df[g] < min_df or len(g) < min_len:
                ignore = True

            # Stoplist.
            elif apply_stoplist:
                for w in g.split():
                    if w in stoplist:
                        ignore = True

            if not ignore:

                # Coerce unicode to string.
                if type(g) is str:
                    g = unicode(g)
                g = unidecode(g)

                if g not in vocab.values():
                    i = len(vocab)
                    vocab[i] = g
                    vocab_[g] = i
                else:
                    i = vocab_[g]
                token_tf[i] += c
                token_df[i] += 1

                t_ngrams[doi].append( (i,c) )

    return t_ngrams, vocab, token_tf

def _handle_paper(article):
    """
    Yields a :class:`.Paper` from an article ET node.

    Parameters
    ----------
    article : Element
        ElementTree Element 'article'.

    Returns
    -------
    paper : :class:`.Paper`
    """
    paper = Paper()
    pdata = dict_from_node(article)

    for key, value in pdata.items():

        datum = pdata[key]
        if type(datum) is str:
            datum = unicode(datum)
        if type(datum) is unicode:
            datum = unidecode(datum).upper()

        paper[key] = datum

    # Handle author names.
    adata = _handle_authors(pdata['author'])
    paper.authors_init = zip(adata[0], adata[1])

    # Handle pubdate.
    paper['date'] = _handle_pubdate(pdata['pubdate'])

    # Handle pagerange.
    paper['spage'], paper['epage'] = _handle_pagerange(pdata['pagerange'])

    return paper

def _handle_pagerange(pagerange):
    """
    Yields start and end pages from DfR pagerange field.

    Parameters
    ----------
    pagerange : str or unicode
        DfR-style pagerange, e.g. "pp. 435-444".

    Returns
    -------
    start : str
        Start page.
    end : str
        End page.
    """

    try:
        pr = re.compile("pp\.\s([0-9]+)\-([0-9]+)")
        start, end = re.findall(pr, pagerange)[0]
    except IndexError:
        start = end = 0

    return unicode(start), unicode(end)

def _handle_pubdate(pubdate):
    """
    Yields a date integer from DfR pubdate field.
    """

    return int(pubdate[0:4])

def _handle_authors(authors):
    """
    Yields aulast and auinit lists from value of authors node.

    Parameters
    ----------
    authors : list, str, or unicode
        Value or values of 'author' element in DfR XML.

    Returns
    -------
    aulast : list
        A list of author surnames (string).
    auinit : list
        A list of author first-initials (string).
    """

    aulast = []
    auinit = []
    if type(authors) is list:
        for author in authors:
            if type(author) is str:
                author = unicode(author)
            author = unidecode(author)
            try:
                l,i = _handle_author(author)
                aulast.append(l)
                auinit.append(i)
            except ValueError:
                pass
    elif type(authors) is str or type(authors) is unicode:
        if type(authors) is str:
            authors = unicode(authors)
        author = unidecode(authors)
        try:
            l,i = _handle_author(author)
            aulast.append(l)
            auinit.append(i)
        except ValueError:
            pass
    else:
        raise ValueError("authors must be a list or a string")

    return aulast, auinit

def _handle_author(author):
    """
    Yields aulast and auinit from an author's full name.

    Parameters
    ----------
    author : str or unicode
        Author fullname, e.g. "Richard L. Nixon".

    Returns
    -------
    aulast : str
        Author surname.
    auinit : str
        Author first-initial.
    """

    lname = author.split(' ')

    try:
        auinit = lname[0][0]
        final = lname[-1].upper()
        if final in ['JR.', 'III']:
            aulast = lname[-2].upper() + " " + final.strip(".")
        else:
            aulast = final
    except IndexError:
        raise ValueError("malformed author name")

    return aulast, auinit

def _dfr2paper_map():
    """
    Defines the direct relationships between DfR article elements and
    :class:`.Paper` fields.

    Returns
    -------
    translator : dict
        A 'translator' dictionary.
    """

    translator = {  'doi': 'doi',
                    'title': 'atitle',
                    'journaltitle': 'jtitle',
                    'volume': 'volume',
                    'issue': 'issue'    }

    return translator

def _create_ayjid(aulast=None, auinit=None, date=None, jtitle=None, **kwargs):
    """
    Convert aulast, auinit, and jtitle into the fuzzy identifier ayjid
    Returns 'Unknown paper' if all id components are missing (None).

    Parameters
    ----------
    Kwargs : dict
        A dictionary of keyword arguments.
    aulast : string
        Author surname.
    auinit: string
        Author initial(s).
    date : string
        Four-digit year.
    jtitle : string
        Title of the journal.

    Returns
    -------
    ayj : string
        Fuzzy identifier ayjid, or 'Unknown paper' if all id components are
        missing (None).

    """
    if aulast is None:
        aulast = ''
    elif isinstance(aulast, list):
        aulast = aulast[0]

    if auinit is None:
        auinit = ''
    elif isinstance(auinit, list):
        auinit = auinit[0]

    if date is None:
        date = ''

    if jtitle is None:
        jtitle = ''

    ayj = aulast + ' ' + auinit + ' ' + unicode(date) + ' ' + jtitle

    if ayj == '   ':
        ayj = 'Unknown paper'

    return ayj.upper()



"""
Methods for building networks from terms in bibliographic records. This
includes keywords, abstract terms, etc.

.. autosummary::
   :nosignatures:

   cooccurrence
   mutual_information
   keyword_cooccurrence
   topic_coupling

"""

from math import log
import networkx as nx
import warnings

from tethne.networks.base import cooccurrence, coupling, multipartite

def _nPMI(p_ij, p_i, p_j):
    lower = (-1.*log(p_ij))
    joint = (p_i*p_j)
    if lower == 0. or joint == 0.:
        return 0.
    return (log(p_ij/joint))/(-1.*log(p_ij))


def feature_cooccurrence(corpus, featureset_name, min_weight=1,
                         filter=lambda f, v, c, dc: True):
    return cooccurrence(corpus, featureset_name, min_weight=min_weight,
                        filter=filter)


def mutual_information(corpus, featureset_name, min_weight=0.9,
                       filter=lambda f, v, c, dc: True):
    """
    Generates a graph of features in ``featureset`` based on normalized
    `pointwise mutual information (nPMI)
    <http://en.wikipedia.org/wiki/Pointwise_mutual_information>`_.

    .. math::

       nPMI(i,j)=\\frac{log(\\frac{p_{ij}}{p_i*p_j})}{-1*log(p_{ij})}

    ...where :math:`p_i` and :math:`p_j` are the probabilities that features
    *i* and *j* will occur in a document (independently), and :math:`p_{ij}` is
    the probability that those two features will occur in the same document.
    """

    graph = feature_cooccurrence(corpus, featureset_name, min_weight=1,
                                 filter=filter)
    mgraph = type(graph)()
    keep_nodes = set()

    fset = corpus.features[featureset_name]
    for s, t, attrs in graph.edges(data=True):
        p_ij = float(attrs['weight'])/len(corpus.papers)
        p_i = float(fset.documentCounts[fset.lookup[s]])/len(corpus.papers)
        p_j = float(fset.documentCounts[fset.lookup[t]])/len(corpus.papers)
        MI = _nPMI(p_ij, p_i, p_j)
        if MI >= min_weight:
            mgraph.add_edge(s, t, nPMI=MI, **attrs)
            keep_nodes.add(s)
            keep_nodes.add(t)

    for n in list(keep_nodes):  # Retain node attributes.
        mgraph.node[n].update(graph.node[n])

    return mgraph

def keyword_cooccurrence(corpus, min_weight=1, filter=lambda f, v, c, dc: True):
    warnings.warn('keyword_cooccurrence will be removed in v0.8. Use ' +
                  'feature_cooccurrence with "authorKeywords" or '+
                  '"keywordsPlus" instead.', DeprecationWarning)
    return feature_cooccurrence(corpus, 'authorKeywords',
                                min_weight=min_weight, filter=filter)



"""
Helper functions.
"""
import string
import copy

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str
    from html.parser import HTMLParser  # Python 3.x
    xrange = range
else:
    from HTMLParser import HTMLParser   # Python 2.x


def is_number(value):
    try:
        int(value)
    except ValueError:
        try:
            float(value)
        except ValueError:
            return False
    return True


def number(value):
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value


def tokenize(s):
    return s.lower()


class MLStripper(HTMLParser):
    def __init__(self):
        super(type(self), self).__init__()

        self.reset()
        self.fed = []
    def handle_data(self, d):
        self.fed.append(d)
    def feed(self, data):
        self.rawdata = self.rawdata + data
        self.goahead(0)
    def get_data(self):
        return u''.join(self.fed)


def strip_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()


def argsort(seq):
    seq = list(seq)
    return sorted(range(len(seq)), key=seq.__getitem__)


def argmin(iterable):
    iterable = list(iterable)
    i_min = -1
    v_min = max(iterable)
    for i, v in enumerate(iterable):
        if v < v_min:
            i_min = i
            v_min = v
    return i_min


def argmax(iterable):
    i_max = -1
    v_max = min(iterable)
    for i, v in enumerate(iterable):
        if v > v_max:
            i_max = i
            v_max = v
    return i_max


def nonzero(iterable):
    return list([i for i, v in enumerate(iterable) if abs(v) > 0.0])


def mean(iterable):
    if len(iterable) > 0:
        return float(sum(iterable))/len(iterable)
    else:
        return float('nan')


def _iterable(o):
    if hasattr(o, '__iter__'):
        return o
    else:
        return [o]


def _strip_punctuation(s):
    """
    Removes all punctuation characters from a string.
    """
    if type(s) is str and not PYTHON_3:    # Bytestring (default in Python 2.x).
        return s.translate(string.maketrans("",""), string.punctuation)
    else:                 # Unicode string (default in Python 3.x).
        translate_table = dict((ord(char), u'') for char in u'!"#%\'()*+,-./:;<=>?@[\]^_`{|}~')
        return s.translate(translate_table)

def _strip_numbers(s):
    """
    Removes all numbers from a string.
    """
    return u''.join([c for c in s if not is_number(c)])


def normalize(s):
    """
    Normalize a token.

    * Convert to lower-case,
    * Remove all punctuation,
    * Remove all numbers.
    """
    return _strip_numbers(_strip_punctuation(s.lower()))


def tokenize(passage):
    """
    Convert a string into a list of normalized words.
    """

    return [normalize(s) for s in passage.split(' ')]


def _space_sep(s):
    if len(s) > 3:
        return s
    return ' '.join(list(s))


def swap(u,v):
    """
    exchange the values of u and v
    """

    return copy.deepcopy(v),copy.deepcopy(u)

def contains(l, f):
    """
    Searches list l for a pattern specified in a lambda function f.
    """
    for x in l:
        if f(x):
            return True
    return False


def overlap(listA, listB):
    """
    Return list of objects shared by listA, listB.
    """
    if (listA is None) or (listB is None):
        return []
    else:
        return list(set(listA) & set(listB))


def subdict(super_dict, keys):
    """
    Returns a subset of the super_dict with the specified keys.
    """
    sub_dict = {}
    valid_keys = super_dict.keys()
    for key in keys:
        if key in valid_keys:
            sub_dict[key] = super_dict[key]

    return sub_dict


def attribs_to_string(attrib_dict, keys):
    """
    A more specific version of the subdict utility aimed at handling
    node and edge attribute dictionaries for NetworkX file formats such as
    gexf (which does not allow attributes to have a list type) by making
    them writable in those formats
    """
    for key, value in attrib_dict.items():
        if (isinstance(value, list) or isinstance(value, dict) or
            isinstance(value, tuple)):
            attrib_dict[key] = value

    return attrib_dict


def concat_list(listA, listB, delim=' '):
    """
    Concatenate list elements pair-wise with the delim character
    Returns the concatenated list
    Raises index error if lists are not parallel
    """

    # Lists must be of equal length.
    if len(listA) != len(listB):
        raise IndexError('Input lists are not parallel.')

    # Concatenate lists.
    listC = []
    for i in xrange(len(listA)):
        app = listA[i] + delim + listB[i]
        listC.append(app)

    return listC

def strip_non_ascii(s):
    """
    Returns the string without non-ASCII characters.

    Parameters
    ----------
    string : string
        A string that may contain non-ASCII characters.

    Returns
    -------
    clean_string : string
        A string that does not contain non-ASCII characters.

    """
    stripped = (c for c in s if 0 < ord(c) < 127)
    clean_string = u''.join(stripped)
    return clean_string

def strip_punctuation(s):
    exclude = set(string.punctuation)
    return u''.join(ch for ch in s if ch not in exclude)


def dict_from_node(node, recursive=False):
    """
    Converts ElementTree node to a dictionary.

    Parameters
    ----------
    node : ElementTree node
    recursive : boolean
        If recursive=False, the value of any field with children will be the
        number of children.

    Returns
    -------
    dict : nested dictionary.
        Tags as keys and values as values. Sub-elements that occur multiple
        times in an element are contained in a list.
    """

    dict = {}
    for snode in node:
        if len(snode) > 0:
            if recursive:
                # Will drill down until len(snode) <= 0.
                value = dict_from_node(snode, True)
            else:
                value = len(snode)
        elif snode.text is not None:
            value = snode.text
        else:
            value = u''

        if snode.tag in dict.keys():    # If there are multiple subelements
                                        #  with the same tag, then the value
                                        #  of the element should be a list
                                        #  rather than a dict.
            if type(dict[snode.tag]) is list:   # If a list has already been
                                                #  started, just append to
                                                #  it.
                dict[snode.tag].append(value)
            else:
                dict[snode.tag] = [ dict[snode.tag], value ]
        else:
            dict[snode.tag] = value     # Default behavior.
    return dict


class Dictionary:
    """
    A two-way index for integer/string pairs.
    """
    def __init__(self):
        self.by_str = {}
        self.by_int = {}

    def __setitem__(self, key, value):
        if type(key) == str:
            self.by_str[key] = value
            self.by_int[value] = key
        if type(key) == int:
            self.by_int[key] = value
            self.by_str[value] = key

    def __getitem__(self, key):
        if type(key) == str:
            return self.by_str[key]
        if type(key) == int:
            return self.by_int[key]



"""
Build networks from topics in a topic model.
"""

import logging
logging.basicConfig(filename=None, format='%(asctime)-6s: %(name)s - %(levelname)s - %(module)s - %(funcName)s - %(lineno)d - %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel('INFO')

import networkx
from math import log

import sys
if sys.version_info[0] > 2:
    xrange = range

from tethne.analyze import features
from tethne.networks.base import cooccurrence, coupling
from tethne.utilities import argsort


def terms(model, threshold=0.01):
    """
    Two terms are coupled if they belong to the same topic with phi > threshold.
    """
    select = lambda f, v, c, dc: v > threshold
    graph = cooccurrence(model.phi, filter=select)

    # Only include labels for terms that are actually in the graph.
    label_map = {k: v for k, v in model.vocabulary.items()
                 if k in graph.nodes()}

    return networkx.relabel_nodes(graph, label_map)


def topic_coupling(model, threshold=None):
    """
    Two papers are coupled if they both contain a shared topic above threshold.
    """
    if not threshold:
        threshold = 3./model.Z
    select = lambda f, v, c, dc: v > threshold

    return coupling(model.corpus, 'topics', filter=select)


def cotopics(model, threshold=None):
    """
    Two topics are coupled if they occur in the same documents.
    """
    if not threshold:
        threshold = 2./model.Z

    select = lambda f, v, c, dc: v > threshold
    return cooccurrence(model.corpus, 'topics', filter=select)


def distance(model, method='cosine', percentile=90, bidirectional=False,
             normalize=True, smooth=False, transform='log'):
    """
    Generate a network of :class:`.Paper`\s based on a distance metric from
    `scipy.spatial.distance
    <http://docs.scipy.org/doc/scipy/reference/spatial.distance.html>`_
    using :ref:`sparse-feature-vector`\s over the dimensions in ``model``.

    Refer to the documentation for :func:`.analyze.features.distance` for
    a list of distance statistics. The only two methods that will not work
    in this context are ``hamming`` and ``jaccard``.

    Distances are inverted to a similarity metric, which is log-transformed by
    default (see ``transform`` parameter, below). Edges are included if they are
    at or above the ``percentile``th percentile.

    Parameters
    ----------
    model : :class:`.LDAModel` or :class:`.DTMModel`
        :func:`.distance` uses ``model.item`` and ``model.metadata``.
    method : str
        Name of a distance method from `scipy.spatial.distance
        <http://docs.scipy.org/doc/scipy/reference/spatial.distance.html>`_.
        See :func:`.analyze.features.distance` for a list of distance
        statistics. ``hamming`` or ``jaccard`` will raise a RuntimeError.
        :func:`.analyze.features.kl_divergence` is also available as
        'kl_divergence'.
    percentile : int
        (default: 90) Edges are included if they are at or above the
        ``percentile`` for all distances in the ``model``.
    bidirectional : bool
        (default: False) If True, ``method`` is calculated twice for each pair
        of :class:`.Paper`\s ( ``(i,j)`` and ``(j,i)`` ), and the mean is used.
    normalize : bool
        (default: True) If True, vectors over topics are normalized so that they
        sum to 1.0 for each :class:`.Paper`.
    smooth : bool
        (default: False) If True, vectors over topics are smoothed according to
        `Bigi 2003
        <http://lvk.cs.msu.su/~bruzz/articles/classification/Using%20Kullback-Leibler%20Distance%20for%20Text%20Categorization.pdf>`_.
        This may be useful if vectors over topics are very sparse.
    transform : str
        (default: 'log') Transformation to apply to similarity values before
        building the graph. So far only 'log' and None are supported.

    Returns
    -------
    thegraph : networkx.Graph
        Similarity values are included as edge weights. Node attributes are set
        using the fields in ``model.metadata``.

    Examples
    --------

    .. code-block:: python

       >>> from tethne.networks import topics
       >>> thegraph = topics.distance(MyLDAModel, 'cosine')

       >>> from tethne.writers import graph
       >>> graph.to_graphml(thegraph, '~./thegraph.graphml')

    .. figure:: _static/images/lda_cosine_network.png
       :width: 80%

       Edge weight and opacity indicate similarity. Node color indicates the
       journal in which each :class:`.Paper` was published. In this graph,
       papers published in the same journal tend to cluster together.

    """

    if method in ['hamming','jaccard']:
        raise RuntimeError(
            'There is no sensicle interpretation of {0} for these data.'
                                                            .format(method))

    thegraph = networkx.Graph()

    edges = {}
    for i in xrange(model.M):
        for j in xrange(i+1, model.M):
            if method == 'kl_divergence':   # Not a SciPy method.
                dist = features.kl_divergence( model.item(i), model.item(j) )
                dist_ = features.kl_divergence( model.item(j), model.item(i) )
                dist = (dist + dist_)/2.
            else:
                dist = features.distance( model.item(i), model.item(j), method,
                                          normalize=normalize, smooth=smooth  )

            if bidirectional:
                dist_ = features.distance(
                            model.item(j), model.item(i), method,
                            normalize=normalize, smooth=smooth  )

                dist = (dist + dist_)/2.

            sim = 1./dist

            if transform == 'log':
                sim = log(sim)

            edges[(i,j)] = sim

    # pct = numpy.percentile(edges.values(), percentile)
    pct = int(round(len(edges)*(percentile/100.)))
    for i in argsort(edge.values())[::-1][:pct]:
        edge, sim = edges.keys()[i], edges.values()[i]
        thegraph.add_edge(edge[0], edge[1], weight=float(sim))

    for key in model.metadata[0].keys():
        values = { k:v[key] for k,v in model.metadata.items()
                                if k in thegraph.nodes()    }
        networkx.set_node_attributes(thegraph, key, values)

    return thegraph



"""
Reader for Scopus CSV data files.
"""

import csv
import re
import os
import uuid

from unidecode import unidecode

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str
    xrange = range

from ..classes import Paper, Corpus

def read_corpus(path):
    """

    """

    papers = read(path)
    return Corpus(papers, index_by='eid')

def corpus_from_dir(path):
    """

    Parameters
    ----------
    path : string
        Path to directory of Scopus CSV data files.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` objects.
    """

    papers = from_dir(path)
    return Corpus(papers, index_by='eid')

def from_dir(path):
    """
    Convenience function for generating a list of :class:`.Paper` from a
    directory of Scopus CSV data files.

    Parameters
    ----------
    path : string
        Path to directory of Scopus CSV data files.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` objects.

    Raises
    ------
    IOError
        Invalid path.

    Examples
    --------

    .. code-block:: python

       >>> from  tethne.readers as rd
       >>> papers = rd.scopus.from_dir("/Path/to/datadir")

    """

    papers = []

    try:
        files = os.listdir(path)
    except IOError:
        raise IOError("Invalid path.")

    for f in files:
        if not f.startswith('.') and f.endswith('csv'): # Ignore hidden files.
            try:
                papers += read('{0}/{1}'.format(path, f))
            except (IOError,UnboundLocalError): # Ignore files that don't
                pass                            #  contain Scopus data.

    return papers

def read(datapath, **kwargs):
    """
    Yields a list of :class:`.Paper` instances from a Scopus CSV data file.

    Parameters
    ----------
    datapath : string
        Filepath to the Web of Science field-tagged data file.

    Returns
    -------
    papers : list
        A list of :class:`.Paper` instances.

    Examples
    --------

    .. code-block:: python

       >>> import tethne.readers as rd
       >>> papers = rd.scopus.read("/Path/to/scopus.csv")

    """

    accession = unicode(uuid.uuid4())   # Accession ID.

    papers = kwargs.get('papers', [])   # Can provide an alternate container.

    rawdata = []
    with open(datapath, 'rb') as f:
        reader = csv.reader(f)
        for row in reader:
            rawdata.append([ unidecode(r.decode('UTF-8')) for r in row ])

    headers = rawdata[0]    # First row is header.
    for datum in rawdata[1:]:
        p = Paper()

        rawdatum = { headers[i]:datum[i] for i in xrange(len(headers)) }
        p['aulast'], p['auinit'] = _handle_authors( rawdatum['Authors'] )
        p['date'] = int(rawdatum['Year'].strip())
        p['ayjid'] = _create_ayjid(
                        p['aulast'], p['auinit'], p['date'], p['jtitle'])

        p['institutions'] = _handle_affiliations(
                rawdatum['Authors with affiliations'], p['aulast'], p['auinit'])
        p['atitle'] = rawdatum['Title'].strip().upper()
        p['jtitle'] = rawdatum['Source title'].strip().upper()
        p['volume'] = rawdatum['Volume'].strip()
        p['issue'] = rawdatum['Issue'].strip()
        p['spage'] = rawdatum['Page start'].strip()
        p['epage'] = rawdatum['Page end'].strip()

        p['abstract'] = rawdatum['Abstract'].strip().upper()
        p['citations'] = _handle_references(rawdatum['References'])
        p['accession'] = accession

        # DOI and PMID are not always present, but must be unique. If they are
        #  not provided, should be set to None so that they can be handled
        #  appropriately downstream. EID should always be present.
        doi = rawdatum['DOI'].strip()
        if doi == '': doi = None
        p['doi'] = doi
        pmid = rawdatum['PubMed ID'].strip()
        if pmid == '': pmid = None
        p['pmid'] = pmid
        p['eid'] = rawdatum['EID'].strip()

        papers.append(p)

    return papers

def _handle_authors(authordata):
    aulast = []
    auinit = []

    for author in authordata.split(', '):
        try:
            a = [a_.strip().upper().replace('.','') for a_ in author.split()]
            aulast.append(' '.join(a[0:-1]))
            auinit.append(a[-1])
        except IndexError:  # Empty record; stray delimiter.
            pass

    return aulast, auinit

def _handle_affiliations(affiliationsdata, aulast, auinit):
    def matching_author(last, init, instcleaned):
        au_parts = [last, init]
        matching = 0
        for x in xrange(len(au_parts)):
            if au_parts[x] == instcleaned[x]:
                matching += 1
        if matching > 1:
            return True, matching
        return False, 0

    institutions = {}

    aff_split = affiliationsdata.split(';')
    A = len(aff_split)
    for i in xrange(A):
        aff = aff_split[i]
        a = aff.split(', ')

        try:
            cleaned = [ a_.strip().upper().replace('.','') for a_ in a ]

            # First work on the assumption that affiliations are in the same
            #  order as author names.
            aul = aulast[i]
            aui = auinit[i]
            match, overlap = matching_author(aul, aui, cleaned)
            if match:
                aname = ' '.join([aul, aui])

            # That might not always be true, so we'll try the other authors
            #  just in case.
            else:
                found = False
                for x in xrange(len(aulast)):
                    aul = aulast[x]
                    aui = auinit[x]

                    match, overlap = matching_author(aul, aui, cleaned)
                    if match:
                        aname = ' '.join([aul, aui])
                        found = True
                        break
                if not found:   # If the author can't be identified, discard.
                    break

            institution = a[match+1:] # The remainder is the institution name.
            l = len(institution)

            if l == 0:
                continue
            if l == 1:
                nation = institution[0].split()[-1].upper()
                inst = ' '.join(institution[0].split()[0:-1]).upper()
            elif l == 2:
                nation = institution[1].upper()
                inst = institution[0].upper()
            else:
                nation = institution[-1].upper()
                inst = ', '.join(institution[0:2]).upper()

            institutions[aname] = [', '.join([inst, nation])]
        except IndexError:  # Blank record part (stray delimiter).
            pass

    # inst_list should be a list of lists, ordered by aulast/auinit.
    authors = [ ' '.join(a) for a in zip(aulast, auinit) ]  # To use as keys.
    inst_list = []
    for au in authors:
        if au in institutions:      # Data available.
            inst_list.append(institutions[au])
        else:
            inst_list.append([])    # No data for that author.

    return inst_list

def _create_ayjid(aulast=None, auinit=None, date=None, jtitle=None, **kwargs):
    """
    Convert aulast, auinit, and jtitle into the fuzzy identifier ayjid
    Returns 'Unknown paper' if all id components are missing (None).

    Parameters
    ----------
    Kwargs : dict
        A dictionary of keyword arguments.
    aulast : string
        Author surname.
    auinit: string
        Author initial(s).
    date : string
        Four-digit year.
    jtitle : string
        Title of the journal.

    Returns
    -------
    ayj : string
        Fuzzy identifier ayjid, or 'Unknown paper' if all id components are
        missing (None).

    """
    if aulast is None or len(aulast) == 0:
        aulast = ''
    elif isinstance(aulast, list):
        aulast = aulast[0]

    if auinit is None or len(auinit) == 0:
        auinit = ''
    elif isinstance(auinit, list):
        auinit = auinit[0]

    if date is None:
        date = ''

    if jtitle is None:
        jtitle = ''

    ayj = aulast + ' ' + auinit + ' ' + unicode(date) + ' ' + jtitle

    if ayj == '   ':
        ayj = 'Unknown paper'

    return ayj.upper()

def _handle_references(referencesdata):
    """
    Use a series of RegEx patterns to (roughly) parse references into
    :class:`.Paper` objects.
    """

    references = [] # A list of Papers.

    for r in referencesdata.strip().split(';'): # Each cited reference.

        # Split at date first, e.g. ' ... (1995) ... '
        m = re.search('(.*)\((?P<date>[0-9]{4})\)(.*)', r)
        if m is not None:
            date = int(m.group('date'))

            pre_date = m.group(1).strip() # Authors, title (unless it's a book).
            post_date = m.group(3).strip() # Journal, volume, pages, etc.

            # Handle authors and title.
            #  Looks for one or more like 'Author, J.K.L., '
            m_ = re.findall('([a-zA-Z]*,\s[A-Z\.]*,[\s]*)', pre_date)

            if m_ is not None:  # Books will pass, but we won't use title.

                # Remove commas between surnames and initials.
                a_ref = ', '.join([ ' '.join(a.split(', ')) for a in m_ ])
                aulast, auinit = _handle_authors(a_ref)

                # Remainder is title.
                l = len(''.join(m_))
                title = pre_date[l:].upper().strip()

            else:   # Let the rest go for now.
                continue

            # Handle journal, volume, issue, pages.
            m_ = re.search('(.*)[p]{1,2}\.(.*)', post_date)

            if m_ is not None:  # Books will fail here.
                pages = m_.group(2).strip()
                pre_pp = m_.group(1)

                # Get start (and end) page number(s).
                if len(pages.split('-')) == 2:  # Both start and end.
                    spage = pages.split('-')[0]
                    epage = pages.split('-')[1]
                else:   # Only one page.
                    spage = pages
                    epage = pages

                # Get journal details.
                jtitle = pre_pp.split(', ')[0].strip().upper()
                issue_volume = pre_pp.split(', ')[1].strip()

                # Look for an issue number, in parentheses.
                j = re.search('([a-z0-9]*) \(([a-z0-9]*)\)', issue_volume)
                if j is not None:
                    volume = j.group(1)
                    issue = j.group(2)

                else:   # No issue number.
                    volume = issue_volume
                    issue = ''

            else:   # Probably a book.
                # TODO: Make this better.
                title = post_date.strip()   # Includes publisher, etc.
                jtitle = ''
                issue = ''
                volume = ''
                spage = ''
                epage = ''

        else:   # Let the rest go for now.
            continue

        # Populate the Paper object.
        p = Paper()
        p['atitle'] = title
        p['aulast'] = aulast
        p['auinit'] = auinit
        p['issue'] = issue
        p['date'] = date
        p['volume'] = volume
        p['jtitle'] = jtitle
        p['spage'] = spage
        p['epage'] = epage
        p['ayjid'] = _create_ayjid(
                            p['aulast'], p['auinit'], p['date'], p['jtitle'])

        references.append(p)

    return references



import networkx as nx
import warnings

from math import exp, log
from collections import defaultdict
from itertools import izip

from tethne.utilities import argmin, mean, argsort


def plot_burstness(corpus, B, **kwargs):
    """
    Generate a figure depicting burstness profiles for ``feature``.

    Parameters
    ----------
    B

    Returns
    -------
    fig : :class:`matplotlib.figure.Figure`

    Examples
    --------

    .. code-block:: python

       >>> from tethne.analyze.corpus import burstness
       >>> fig = plot_burstness(corpus, 'citations', topn=2, perslice=True)
       >>> fig.savefig('~/burstness.png')

    Years prior to the first occurrence of each feature are grayed out. Periods
    in which the feature was bursty are depicted by colored blocks, the opacity
    of which indicates burstness intensity.

    .. figure:: _static/images/burstness.png
       :width: 600
       :align: center
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches
    except ImportError:
        raise RuntimeError('This method requires the package matplotlib.')

    color = kwargs.get('color', 'red')

    # Get width based on slices.
    years = sorted(corpus.indices['date'].keys())
    width = years[1] - years[0]
    height = 1.0

    fig = plt.figure(figsize=(10,len(B)/4.))

    f = 1
    axes = {}
    for key, value in B.iteritems():
        x,y = value
        ax = fig.add_subplot(len(B),1,f)
        f+=1
        ax.set_yticks([])
        ax.set_xbound(min(years), max(years) + 1)

        if not f == len(B)+1:   # Only show xticks on the bottom subplot.
            ax.set_xticklabels([])

        # Block out years until first occurrence of feature.

        rect = mpatches.Rectangle((min(years), 0), sorted(x)[0]-min(years),
                                   height, fill=True, linewidth=0.0)
        rect.set_facecolor('black')
        rect.set_alpha(0.3)
        ax.add_patch(rect)

        # Add a rectangle for each year, shaded according to burstness state.
        for d in xrange(min(x), max(x)):
            try:
                i = x.index(d)
            except ValueError:
                continue

            xy = (d, 0.)

            state = y[i]
            rect = mpatches.Rectangle(xy, width, height,
                                      fill=True, linewidth=0.0)
            rect.set_facecolor(color)
            rect.set_alpha(state)
            ax.add_patch(rect)

        ax.set_ylabel(key, rotation=0,
                      horizontalalignment='right',
                      verticalalignment='center')
    plt.subplots_adjust(left=0.5)
    fig.tight_layout(h_pad=0.25)

    plt.show()


def plot_sigma(corpus, sigma, nodes=None, **kwargs):
    """
    Plot sigma values for the ``topn`` most influential nodes.

    Parameters
    ----------
    G : :class:`.GraphCollection`
    corpus : :class:`.Corpus`
    feature : str
        Name of a featureset in `corpus`.
    topn : int or float {0.-1.}
        (default: 20) Number (int) or percentage (float) of top-occurring
        features to return. If ``flist`` is provided, this parameter is ignored.
    sort_by : str
        (default: 'max') Criterion for selecting ``topn`` nodes.
    perslice : bool
        (default: False) If True, loads ``topn`` features per slice. Otherwise,
        loads ``topn`` features overall. If ``flist`` is provided, this
        parameter is ignored.
    flist : list
        List of nodes. If provided, ``topn`` and ``perslice`` are ignored.
    fig : :class:`matplotlib.figure.Figure`
        (default: None) You may provide a Figure instance if you wish.
        Otherwise, a new figure is generated.

    Returns
    -------
    fig : :class:`matplotlib.figure.Figure`
    G : :class:`.GraphCollection`
        A co-citation graph collection, updated with ``sigma`` node attributes.

    Examples
    --------

    Assuming that you have a :class:`.Corpus` (``G``) sliced by ``'date'`` and a
    co-citation :class:`.GraphCollection` (``corpus``)...

    .. code-block:: python

       >>> from tethne.analyze.cocitation import plot_sigma
       >>> fig,G = plot_sigma(G, corpus, topn=5, perslice=True)
       >>> fig.savefig('~/sigma_plot.png')

    In this figure, the top 5 most sigma-influential nodes in each slice are
    shown. Red bands indicate periods in which each paper was influential;
    opacity indicates the intensity of sigma (normalized by the highest value in
    the plot). The period prior to the first instance of each node is grayed
    out.

    .. figure:: _static/images/sigma_plot.png
       :width: 600
       :align: center
    """

    try:
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches
    except ImportError:
        raise RuntimeError('This method requires the package matplotlib.')

    if nodes == 'all':
        nodes = sigma.keys()

    # Display parameters.
    color = kwargs.get('color', 'red')
    years = sorted(corpus.indices['date'].keys())
    width = years[1] - years[0] # Get width based on slices.
    height = 1.0

    sort_by = kwargs.get('sort_by', 'max')
    perslice = kwargs.get('perslice', False)
    topn = kwargs.get('topn', 20)

    if not nodes:
        # Get only the topn most significant papers.
        include = []
        if sort_by == 'max':
            if perslice:    # Get topn per slice.
                vals = {}
                norm_by = 0.

                # Organize values in a way that makes selection easier.
                for node, history in sigma.iteritems():
                    years, values = history
                    if max(values) == 0.:
                        continue
                    for year, val in zip(years, values):
                        if year not in vals:
                            vals[year] = {}
                        vals[year][node] = val

                # Get the maximum values for each slice.
                for year, node_values in vals.iteritems():
                    indices = argsort(node_values.values())[-topn:][::-1]
                    include += [node_values.keys()[i] for i in indices]
                    max_value = max(node_values.values())
                    if max_value > norm_by:
                        norm_by = max_value

            else:   # Get topn overall.
                maxes = [max(v[1]) for v in sigma.values() ]
                indices = argsort(maxes)[-topn:][::-1]
                include = [sigma.keys()[i] for i in indices]
                norm_by = max(maxes)

        # Nodes to include.
        nodes = [node for node, values in sigma.iteritems()
                 if max(values[1]) > 0 and node in include]

#     if fig is None: # Create a new Figure instance.
    fig = plt.figure(figsize=(10, len(nodes)/4.))

    # Plot!
    f = 1   # Current subplot.
    axes = {}

    # Earliest year for which we have values.
    x_min = min([min(years) for years, values in sigma.values()])

    for node in nodes:
        x_order = argsort(sigma[node][0])
        x = sorted(sigma[node][0])
        y = [sigma[node][1][i]/norm_by for i in x_order]

        ax = fig.add_subplot(len(nodes), 1, f)
        f+=1
        ax.set_yticks([])
        ax.set_xbound(x_min, max(years)+1)

        # Only show xticks on the bottom subplot.
        if not f == len(nodes) + 1:
            ax.set_xticklabels([])

        # Block out years until first occurrence of feature.
        rect = mpatches.Rectangle((x_min, 0), x[0] - x_min,
                                  height, fill=True, linewidth=0.0)
        rect.set_facecolor('black')
        rect.set_alpha(0.1)
        ax.add_patch(rect)

        # Add a rectangle for each year, shaded according to burstness state.
        for d in xrange(min(x), max(x)):
            try:    # May not have values for all years.
                i = x.index(d)
            except ValueError:
                continue

            xy = (d, 0.)

            state = y[i]
            rect = mpatches.Rectangle(xy, width, height,
                                      fill=True, linewidth=0.0)
            rect.set_facecolor(color)
            rect.set_alpha(state + 0.1)
            ax.add_patch(rect)

        ax.set_ylabel(node, rotation=0,
                      horizontalalignment='right',
                      verticalalignment='center')

    plt.subplots_adjust(left=0.5)
    fig.tight_layout(h_pad=0.25)
    plt.show()



import os
import re
import xml.etree.ElementTree as ET
import rdflib

import codecs
import chardet
import unicodedata

import logging

# rdflib complains a lot.
logging.getLogger("rdflib").setLevel(logging.ERROR)

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


class dobject(object):
    pass


def _cast(value):
    """
    Attempt to convert ``value`` to an ``int`` or ``float``. If unable, return
    the value unchanged.
    """

    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value


class BaseParser(object):
    """
    Base class for all data parsers. Do not instantiate directly.
    """
    def __init__(self, path, **kwargs):
        self.path = path
        self.data = []
        self.fields = set([])

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.open()

    def new_entry(self):
        """
        Prepare a new data entry.
        """
        self.data.append(self.entry_class())

    def _get_handler(self, tag):
        handler_name = 'handle_{tag}'.format(tag=tag)
        if hasattr(self, handler_name):
            return getattr(self, handler_name)
        return

    def set_value(self, tag, value):
        setattr(self.data[-1], tag, value)

    def postprocess_entry(self):
        for field in self.fields:
            processor_name = 'postprocess_{0}'.format(field)
            if hasattr(self.data[-1], field) and hasattr(self, processor_name):
                getattr(self, processor_name)(self.data[-1])


class IterParser(BaseParser):
    entry_class = dobject
    """Model for data entry."""

    concat_fields = []
    """
    Multi-line fields here should be concatenated, rather than represented
    as lists.
    """

    tags = {}

    def __init__(self, *args, **kwargs):
        super(IterParser, self).__init__(*args, **kwargs)

        self.current_tag = None
        self.last_tag = None

        if kwargs.get('autostart', True):
            self.start()

    def parse(self):
        """

        """
        while True:        # Main loop.
            tag, data = self.next()
            if self.is_eof(tag):
                self.postprocess_entry()
                break

            self.handle(tag, data)
            self.last_tag = tag
        return self.data

    def start(self):
        """
        Find the first data entry and prepare to parse.
        """

        while not self.is_start(self.current_tag):
            self.next()
        self.new_entry()

    def handle(self, tag, data):
        """
        Process a single line of data, and store the result.

        Parameters
        ----------
        tag : str
        data :
        """
        if isinstance(data,unicode):
            data = unicodedata.normalize('NFKD', data)#.encode('utf-8','ignore')

        if self.is_end(tag):
            self.postprocess_entry()

        if self.is_start(tag):
            self.new_entry()

        if data is None or tag is None:
            return

        handler = self._get_handler(tag)
        if handler is not None:
            data = handler(data)

        if tag in self.tags:    # Rename the field.
            tag = self.tags[tag]

        # Multiline fields are represented as lists of values.
        if hasattr(self.data[-1], tag):
            value = getattr(self.data[-1], tag)
            if tag in self.concat_fields:
                value = ' '.join([value, data])
            elif type(value) is list:
                value.append(data)
            elif value not in [None, '']:
                value = [value, data]
        else:
            value = data
        setattr(self.data[-1], tag, value)
        self.fields.add(tag)


class FTParser(IterParser):
    """
    Base parser for field-tagged data files.
    """

    start_tag = 'ST'
    """Signals the start of a data entry."""

    end_tag = 'ED'
    """Signals the end of a data entry."""

    def is_start(self, tag):
        return tag == self.start_tag

    def is_end(self, tag):
        return tag == self.end_tag

    def is_eof(self, tag):
        return self.at_eof

    def open(self):
        """
        Open the data file.
        """

        if not os.path.exists(self.path):
            raise IOError("No such path: {0}".format(self.path))


        with open(self.path, "rb") as f:
            msg = f.read()
        result = chardet.detect(msg)

        self.buffer = codecs.open(self.path, "rb", encoding=result['encoding'])

        self.at_eof = False

    def next(self):
        """
        Get the next line of data.

        Returns
        -------
        tag : str
        data :
        """
        line = self.buffer.readline()

        while line == '\n':       # Skip forward to the next line with content.
            line = self.buffer.readline()

        if line == '':            # End of file.
            self.at_eof = True
            return None, None

        match = re.match('([A-Z]{2})\W+(.*)', line)
        if match is not None:
            self.current_tag, data = match.groups()
        else:
            self.current_tag = self.last_tag
            data = line.strip()
        return self.current_tag, _cast(data)

    def __del__(self):
        if hasattr(self, 'buffer'):
            self.buffer.close()


class XMLParser(IterParser):
    entry_element = 'article'
    entry_class = dobject

    def open(self):
        with open(self.path, 'r') as f:
            self.root = ET.fromstring(f.read())
        pattern = './/{elem}'.format(elem=self.entry_element)
        self.elements = self.root.findall(pattern)

        self.at_start = False
        self.at_end = False
        self.children = []

    def is_start(self, tag):
        return self.at_start

    def is_end(self, tag):
        return self.at_end

    def is_eof(self, tag):
        return len(self.elements) == 0 and len(self.children) == 0

    def next(self):
        if len(self.children) > 0:
            child = self.children.pop(0)
            tag, text = child.tag, child.text

        else:
            tag, text = None, None
            if self.at_end:
                self.at_end = False
                self.at_start = True

            elif self.at_start:
                element = self.elements.pop(0)
                self.children = element.getchildren()
                self.at_start = False
                tag, text = self.next()

            elif len(self.data) == 0 and len(self.elements) > 0:
                self.at_start = True
            else:
                self.at_end = True

        return tag, text


class RDFParser(BaseParser):
    entry_elements = ['Document']   #
    meta_elements = []
    concat_fields = []

    def open(self):
        self.graph = rdflib.Graph()
        self.graph.parse(self.path)
        self.entries = []

        for element in self.entry_elements:
            query = 'SELECT * WHERE { ?p a ' + element + ' }'
            self.entries += [r[0] for r in self.graph.query(query)]

    def next(self):
        if len(self.entries) > 0:
            return self.entries.pop(0)

    def parse(self):
        meta_fields, meta_refs = zip(*self.meta_elements)

        while True:        # Main loop.
            entry = self.next()
            if entry is None:
                break

            self.new_entry()

            for s, p, o in self.graph.triples((entry, None, None)):
                if p in meta_refs:  # Look for metadata fields.
                    tag = meta_fields[meta_refs.index(p)]
                    self.handle(tag, o)
            self.postprocess_entry()

        return self.data

    def handle(self, tag, data):
        handler = self._get_handler(tag)

        if handler is not None:
            data = handler(data)

        if tag in self.tags:    # Rename the field.
            tag = self.tags[tag]

        if data is not None:
            # Multiline fields are represented as lists of values.
            if hasattr(self.data[-1], tag):
                value = getattr(self.data[-1], tag)
                if tag in self.concat_fields:
                    value = ' '.join([value, data])
                elif type(value) is list:
                    value.append(data)
                elif value not in [None, '']:
                    value = [value, data]
            else:
                value = data

            setattr(self.data[-1], tag, value)
            self.fields.add(tag)
